""" Add CLI support to start MIP rare disease RNA"""

import logging

import click

from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI

from cg.apps.environ import environ_email
from cg.cli.workflow.get_links import get_links
from cg.cli.workflow.mip.store import store as store_cmd
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.constants import Pipeline
from cg.store import Store
from cg.store.utils import case_exists

LOG = logging.getLogger(__name__)


@click.group("mip-rna")
@click.pass_context
def mip_rna(context: click.Context):
    """Rare disease RNA workflow"""
    context.obj["housekeeper_api"] = HousekeeperAPI(context.obj)
    context.obj["trailblazer_api"] = TrailblazerAPI(context.obj)
    context.obj["scout_api"] = ScoutAPI(context.obj)
    context.obj["lims_api"] = LimsAPI(context.obj)
    context.obj["status_db"] = Store(context.obj["database"])

    context.obj["rna_api"] = MipAnalysisAPI(
        db=context.obj["status_db"],
        hk_api=context.obj["housekeeper_api"],
        tb_api=context.obj["trailblazer_api"],
        scout_api=context.obj["scout_api"],
        lims_api=context.obj["lims_api"],
        script=context.obj["mip-rd-rna"]["script"],
        pipeline=context.obj["mip-rd-rna"]["pipeline"],
        conda_env=context.obj["mip-rd-rna"]["conda_env"],
        root=context.obj["mip-rd-rna"]["root"],
    )


@mip_rna.command()
@click.option("-c", "--case", "case_id", help="link all samples for a case")
@click.argument("sample_id", required=False)
@click.pass_context
def link(context: click.Context, case_id: str, sample_id: str):
    """Link FASTQ files for a sample_id"""
    rna_api = context.obj["rna_api"]
    link_objs = get_links(rna_api.db, case_id, sample_id)

    for link_obj in link_objs:
        LOG.info(
            "%s: %s link FASTQ files", link_obj.sample.internal_id, link_obj.family.data_analysis
        )
        if "mip + rna" in link_obj.family.data_analysis.lower():
            rna_api.link_sample(
                sample=link_obj.sample,
                case_id=case_id,
            )


@mip_rna.command()
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.option("-e", "--email", help="email to send errors to")
@click.option("--mip-dry-run", "mip_dry_run", is_flag=True, help="Run MIP in dry-run mode")
@click.option("-p", "--priority", type=click.Choice(["low", "normal", "high"]))
@click.option("-sw", "--start-with", help="start mip from this program.")
@click.argument("case_id")
@click.pass_context
def run(
    context: click.Context,
    case_id: str,
    dry_run: bool = False,
    mip_dry_run: bool = False,
    priority: str = None,
    email: str = None,
    start_with: str = None,
):
    """Run the analysis for a case"""
    rna_api = context.obj["rna_api"]
    case_obj = rna_api.db.family(case_id)

    if not case_exists(case_obj, case_id):
        raise click.Abort()
    if rna_api.get_analyses_from_trailblazer(case_id=case_obj.internal_id, temp=True):
        LOG.warning("%s: analysis already running", case_obj.internal_id)
        return

    email = email or environ_email()
    kwargs = dict(
        config=context.obj["mip-rd-rna"]["mip_config"],
        case=case_id,
        priority=priority,
        email=email,
        dryrun=mip_dry_run,
        start_with=start_with,
    )
    if dry_run:
        rna_api.run_command(dry_run=dry_run, **kwargs)
        return

    rna_api.run_command(**kwargs)

    if mip_dry_run:
        LOG.info("Executed MIP in dry-run mode - skipping Trailblazer step")
        return

    rna_api.mark_analyses_deleted(case_id=case_id)
    rna_api.add_pending_analysis(
        case_id=case_id,
        email=email,
        type=rna_api.get_application_type(case_id),
        out_dir=rna_api.get_case_output_path(case_id).as_posix(),
        config_path=rna_api.get_slurm_job_ids_path(case_id).as_posix(),
        priority="normal",
        data_analysis=Pipeline.MIP_RNA,
    )
    rna_api.set_statusdb_action(case_id=case_id, action="running")
    LOG.info("MIP rd-rna run started!")


@mip_rna.command("config-case")
@click.option("-d", "--dry", is_flag=True, help="Print config to console")
@click.argument("case_id")
@click.pass_context
def config_case(context: click.Context, case_id: str, dry: bool = False):
    """Generate a config for the case_id"""
    rna_api = context.obj["rna_api"]

    case_obj = rna_api.db.family(case_id)
    if not case_exists(case_obj, case_id):
        context.abort()
    config_data = rna_api.pedigree_config(case_obj, pipeline=Pipeline.MIP_RNA)
    if dry:
        print(config_data)
        return
    out_path = rna_api.write_pedigree_config(config_data)
    LOG.info(f"Config saved to: {out_path}")


mip_rna.add_command(store_cmd)



"""Unified interface to handle sample submissions.

This interface will update information in Status and/or LIMS as required.

The normal entry for information is through the REST API which will pass a JSON
document with all information about samples in the submission. The input will
be validated and if passing all checks be accepted as new samples.
"""
import datetime as dt
import logging
import re
from typing import List

import typing
from cg.apps.lims import LimsAPI
from cg.apps.osticket import OsTicket
from cg.constants import Pipeline
from cg.exc import OrderError, TicketCreationError
from cg.store import Store, models
from .schema import OrderType, ORDER_SCHEMES
from .lims import LimsHandler
from .status import StatusHandler

LOG = logging.getLogger(__name__)


class OrdersAPI(LimsHandler, StatusHandler):
    """Orders API for accepting new samples into the system."""

    def __init__(self, lims: LimsAPI, status: Store, osticket: OsTicket = None):
        super().__init__()
        self.lims = lims
        self.status = status
        self.osticket = osticket

    def submit(self, project: OrderType, data: dict, ticket: dict) -> dict:
        """Submit a batch of samples.

        Main entry point for the class towards interfaces that implements it.
        """
        try:
            ORDER_SCHEMES[project].validate(data)
        except (ValueError, TypeError) as error:
            raise OrderError(error.args[0])

        self._validate_customer_on_imported_samples(project, data)

        # detect manual ticket assignment
        ticket_match = re.fullmatch(r"#([0-9]{6})", data["name"])

        if ticket_match:
            ticket_number = int(ticket_match.group(1))
            LOG.info(f"{ticket_number}: detected ticket in order name")
            data["ticket"] = ticket_number
        else:
            # open and assign ticket to order
            try:
                if self.osticket:
                    message = f"data:text/html;charset=utf-8,New incoming samples: "

                    for sample in data.get("samples"):
                        message += "<br />" + sample.get("name")

                        if sample.get("application"):
                            message += f", application: {sample.get('application')}"

                        if sample.get("family_name"):
                            message += f", family: {sample.get('family_name')}"

                        if sample.get("internal_id"):

                            existing_sample = self.status.sample(sample.get("internal_id"))
                            sample_customer = ""
                            if existing_sample.customer_id != data["customer"]:
                                sample_customer = " from " + existing_sample.customer.internal_id

                            message += f" (already existing sample{sample_customer})"

                        if sample.get("comment"):
                            message += ", " + sample.get("comment")

                    message += f"<br />"

                    if data.get("comment"):
                        message += f"<br />{data.get('comment')}."

                    if ticket.get("name"):
                        message += f"<br />{ticket.get('name')}"

                    data["ticket"] = self.osticket.open_ticket(
                        name=ticket["name"],
                        email=ticket["email"],
                        subject=data["name"],
                        message=message,
                    )

                    LOG.info(f"{data['ticket']}: opened new ticket")
                else:
                    data["ticket"] = None
            except TicketCreationError as error:
                LOG.warning(error.message)
                data["ticket"] = None
        order_func = self._get_submit_func(project.value)
        result = order_func(data)
        return result

    def submit_rml(self, data: dict) -> dict:
        """Submit a batch of ready made libraries."""
        status_data = self.pools_to_status(data)
        project_data, _ = self.process_lims(data, data["samples"])
        new_records = self.store_pools(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket=data["ticket"],
            pools=status_data["pools"],
        )
        return {"project": project_data, "records": new_records}

    def submit_fastq(self, data: dict) -> dict:
        """Submit a batch of samples for FASTQ delivery."""
        status_data = self.samples_to_status(data)
        project_data, lims_map = self.process_lims(data, data["samples"])
        self.fill_in_sample_ids(status_data["samples"], lims_map)
        new_samples = self.store_fastq_samples(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket=data["ticket"],
            samples=status_data["samples"],
        )
        self.add_missing_reads(new_samples)
        return {"project": project_data, "records": new_samples}

    def submit_metagenome(self, data: dict) -> dict:
        """Submit a batch of metagenome samples."""
        status_data = self.samples_to_status(data)
        project_data, lims_map = self.process_lims(data, data["samples"])
        self.fill_in_sample_ids(status_data["samples"], lims_map)
        new_samples = self.store_samples(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket=data["ticket"],
            samples=status_data["samples"],
        )
        self.add_missing_reads(new_samples)
        return {"project": project_data, "records": new_samples}

    def submit_external(self, data: dict) -> dict:
        """Submit a batch of externally sequenced samples for analysis."""
        result = self.process_family_samples(data)
        return result

    def submit_case_samples(self, data: dict) -> dict:
        """Submit a batch of samples for sequencing and analysis."""
        result = self.process_family_samples(data)
        for family_obj in result["records"]:
            LOG.info(f"{family_obj.name}: submit family samples")
            status_samples = [
                link_obj.sample
                for link_obj in family_obj.links
                if link_obj.sample.ticket_number == data["ticket"]
            ]
            self.add_missing_reads(status_samples)
        self.update_application(data["ticket"], result["records"])
        return result

    def submit_mip_dna(self, data: dict) -> dict:
        """Submit a batch of samples for sequencing and analysis."""
        return self.submit_case_samples(data)

    def submit_balsamic(self, data: dict) -> dict:
        """Submit a batch of samples for sequencing and balsamic analysis."""
        return self.submit_case_samples(data)

    def submit_mip_rna(self, data: dict) -> dict:
        """Submit a batch of samples for sequencing and analysis."""
        return self.submit_case_samples(data)

    def submit_microsalt(self, data: dict) -> dict:
        """Submit a batch of microbial samples."""
        # prepare data for status database
        status_data = self.microbial_samples_to_status(data)
        self.fill_in_sample_verified_organism(data["samples"])
        # submit samples to LIMS
        project_data, lims_map = self.process_lims(data, data["samples"])
        self.fill_in_sample_ids(status_data["samples"], lims_map, id_key="internal_id")
        # submit samples to Status
        samples = self.store_microbial_samples(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"] if project_data else dt.datetime.now(),
            ticket=data["ticket"],
            samples=status_data["samples"],
            comment=status_data["comment"],
            data_analysis=Pipeline(status_data["data_analysis"]),
        )

        return {"project": project_data, "records": samples}

    def process_family_samples(self, data: dict) -> dict:
        """Process samples to be analyzed."""
        # filter out only new samples
        status_data = self.cases_to_status(data)
        new_samples = [sample for sample in data["samples"] if sample.get("internal_id") is None]
        if new_samples:
            project_data, lims_map = self.process_lims(data, new_samples)
        else:
            project_data = lims_map = None
        samples = [sample for family in status_data["families"] for sample in family["samples"]]
        if lims_map:
            self.fill_in_sample_ids(samples, lims_map)
        new_families = self.store_cases(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"] if project_data else dt.datetime.now(),
            ticket=data["ticket"],
            cases=status_data["families"],
        )
        return {"project": project_data, "records": new_families}

    def update_application(self, ticket_number: int, families: List[models.Family]) -> None:
        """Update application for trios if relevant."""
        reduced_map = {
            "EXOSXTR100": "EXTSXTR100",
            "WGSPCFC030": "WGTPCFC030",
        }
        for case_obj in families:
            LOG.debug(f"{case_obj.name}: update application for trios")
            order_samples = [
                link_obj.sample
                for link_obj in case_obj.links
                if link_obj.sample.ticket_number == ticket_number
            ]
            if len(order_samples) >= 3:
                applications = [
                    sample_obj.application_version.application for sample_obj in order_samples
                ]
                prep_categories = set(application.prep_category for application in applications)
                if len(prep_categories) == 1:
                    for sample_obj in order_samples:
                        if not sample_obj.application_version.application.reduced_price:
                            application_tag = sample_obj.application_version.application.tag
                            if application_tag in reduced_map:
                                reduced_tag = reduced_map[application_tag]
                                LOG.info(
                                    f"{sample_obj.internal_id}: update application tag - "
                                    f"{reduced_tag}"
                                )
                                reduced_version = self.status.current_application_version(
                                    reduced_tag
                                )
                                sample_obj.application_version = reduced_version

    def add_missing_reads(self, samples: List[models.Sample]):
        """Add expected reads/reads missing."""
        for sample_obj in samples:
            LOG.info(f"{sample_obj.internal_id}: add missing reads in LIMS")
            target_reads = sample_obj.application_version.application.target_reads / 1000000
            self.lims.update_sample(sample_obj.internal_id, target_reads=target_reads)

    @staticmethod
    def fill_in_sample_ids(samples: List[dict], lims_map: dict, id_key: str = "internal_id"):
        """Fill in LIMS sample ids."""
        for sample in samples:
            LOG.debug(f"{sample['name']}: link sample to LIMS")
            if not sample.get(id_key):
                internal_id = lims_map[sample["name"]]
                LOG.info(f"{sample['name']} -> {internal_id}: connect sample to LIMS")
                sample[id_key] = internal_id

    def fill_in_sample_verified_organism(self, samples: List[dict]):
        for sample in samples:
            organism_id = sample["organism"]
            reference_genome = sample["reference_genome"]
            organism = self.status.organism(internal_id=organism_id)
            is_verified = (
                organism and organism.reference_genome == reference_genome and organism.verified
            )
            sample["verified_organism"] = is_verified

    def _validate_customer_on_imported_samples(self, project, data):
        for sample in data.get("samples"):

            if sample.get("internal_id"):

                if project not in (
                    OrderType.BALSAMIC,
                    OrderType.EXTERNAL,
                    OrderType.MIP_DNA,
                    OrderType.MIP_RNA,
                ):
                    raise OrderError(
                        f"Only MIP, Balsamic and external orders can have imported "
                        f"samples: "
                        f"{sample.get('name')}"
                    )

                existing_sample = self.status.sample(sample.get("internal_id"))
                data_customer = self.status.customer(data["customer"])

                if existing_sample.customer.customer_group_id != data_customer.customer_group_id:
                    raise OrderError(f"Sample not available: {sample.get('name')}")

    def _get_submit_func(self, project_type: OrderType) -> typing.Callable:
        """Get the submit method to call for the given type of project"""

        if project_type == OrderType.MIP_DNA:
            return getattr(self, "submit_mip_dna")
        elif project_type == OrderType.MIP_RNA:
            return getattr(self, "submit_mip_rna")

        return getattr(self, f"submit_{str(project_type)}")



""" Test the __getattr__ override when calling private _store """

import logging


def test_calling_method_on_private_store_give_warning(housekeeper_api, caplog):
    """Test that we get a log warning for unwrapped methods"""

    # GIVEN an hk api and a method that is not wrapped
    caplog.set_level(logging.WARNING)

    # WHEN we call add_file
    housekeeper_api.files_before()

    # THEN the log should contain a warning that we have called something non-wrapped
    with caplog.at_level(logging.WARNING):
        assert "files_before" in caplog.text
        assert "HousekeeperAPI" in caplog.text



"""Code for uploading to scout via CLI"""
import logging
from pathlib import Path

import click
import yaml

from cg.apps.hk import HousekeeperAPI
from cg.meta.upload.scoutapi import UploadScoutAPI

from .utils import suggest_cases_to_upload

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-r", "--re-upload", is_flag=True, help="re-upload existing analysis")
@click.option("-p", "--print", "print_console", is_flag=True, help="print config values")
@click.argument("case_id", required=False)
@click.pass_context
def scout(context, re_upload, print_console, case_id):
    """Upload variants from analysis to Scout."""

    click.echo(click.style("----------------- SCOUT -----------------------"))

    if not case_id:
        suggest_cases_to_upload(context)
        context.abort()

    status_api = context.obj["status_db"]
    scout_upload_api = context.obj["scout_upload_api"]
    hk_api = context.obj["housekeeper_api"]
    family_obj = status_api.family(case_id)
    scout_config = scout_upload_api.generate_config(family_obj.analyses[0])
    mip_dna_root_dir = context.obj["mip-rd-dna"]["root"]

    if print_console:
        click.echo(scout_config)
        return

    file_path = Path(mip_dna_root_dir, case_id, "scout_load.yaml")

    if file_path.exists():
        message = (
            "Scout load config %s already exists, you might remove the file and try "
            "again, consider that you might also have it in housekeeper" % file_path
        )
        LOG.warning(message)
        context.abort()

    scout_upload_api.save_config_file(scout_config, file_path)
    try:
        LOG.info("Upload file to housekeeper: %s", file_path)
        scout_upload_api.add_scout_config_to_hk(file_path, hk_api, case_id)
    except FileExistsError as err:
        LOG.warning("%s, consider removing the file from housekeeper and try again", str(err))
        context.abort()

    context.invoke(upload_case_to_scout, case_id=case_id, re_upload=re_upload)


@click.command()
@click.option("-r", "--re-upload", is_flag=True, help="re-upload existing analysis")
@click.option("--dry-run", is_flag=True)
@click.argument("case_id")
@click.pass_context
def upload_case_to_scout(context, re_upload, dry_run, case_id):
    """Upload variants and case from analysis to Scout."""

    click.echo(click.style("----------------- CONFIG -----------------------"))

    def _get_load_config_from_hk(hk_api: HousekeeperAPI, case_id):
        tag_name = UploadScoutAPI.get_load_config_tag()
        version_obj = hk_api.last_version(case_id)
        scout_config_files = hk_api.get_files(
            bundle=case_id, tags=[tag_name], version=version_obj.id
        )
        if len(list(scout_config_files)) == 0:
            raise FileNotFoundError(f"No scout load config was found in housekeeper for {case_id}")

        return scout_config_files[0].full_path

    scout_api = context.obj["scout_api"]
    hk_api = context.obj["housekeeper_api"]

    load_config = _get_load_config_from_hk(hk_api, case_id)

    LOG.info("uploading case %s to scout", case_id)
    with open(load_config, "r") as stream:
        scout_configs = yaml.safe_load(stream)

    if not dry_run:
        scout_api.upload(scout_configs, force=re_upload)

    click.echo(
        click.style("uploaded to scout using load config {}".format(load_config), fg="green")
    )



"""Constants for delivery"""


BALSAMIC_ANALYSIS_CASE_TAGS = [
    {"cnv", "annotated-somatic-vcf-all", "vcf-all", "cnvkit"},
    {"cnv", "annotated-somatic-vcf-all-index", "vcf-all", "cnvkit"},
    {"snv", "annotated-somatic-vcf-pass", "vcf-pass", "vardict"},
    {"snv", "annotated-somatic-vcf-pass-index", "vcf-pass", "vardict"},
    {"snv", "annotated-somatic-vcf-pass", "vcf-pass", "tnscope"},
    {"snv", "annotated-somatic-vcf-pass-index", "vcf-pass", "tnscope"},
    {"snv", "annotated-somatic-vcf-pass", "vcf-pass", "tnhaplotyper"},
    {"snv", "annotated-somatic-vcf-pass-index", "vcf-pass", "tnhaplotyper"},
    {"sv", "annotated-somatic-vcf-pass", "vcf-pass", "manta"},
    {"sv", "annotated-somatic-vcf-pass-index", "vcf-pass", "manta"},
    {"cnv-scatter"},
    {"cnv-diagram"},
    {"multiqc-html"},
    {"balsamic-report"},
    {"normal-cram"},
    {"normal-cram-index"},
    {"tumor-cram"},
    {"tumor-cram-index"},
]
BALSAMIC_ANALYSIS_SAMPLE_TAGS = [
    {"normal-cram"},
    {"normal-cram-index"},
    {"tumor-cram"},
    {"tumor-cram-index"},
    {"bam"},
    {"quality-trimmed-fastq-read1"},
    {"quality-trimmed-fastq-read2"},
]

BALSAMIC_QC_CASE_TAGS = [
    {"multiqc-html"},
]
BALSAMIC_QC_SAMPLE_TAGS = [
    {"quality-trimmed-fastq-read1"},
    {"quality-trimmed-fastq-read2"},
]


MIP_DNA_ANALYSIS_CASE_TAGS = [
    {"vcf-clinical-sv-bin"},
    {"vcf-clinical-sv-bin-index"},
    {"vcf-research-sv-bin"},
    {"vcf-research-sv-bin-index"},
    {"gbcf"},
    {"gbcf-index"},
    {"snv-gbcf"},
    {"snv-gbcf-index"},
    {"snv-bcf"},
    {"snv-bcf-index"},
    {"sv-bcf"},
    {"sv-bcf-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
    {"vcf-sv-clinical"},
    {"vcf-sv-clinical-index"},
    {"vcf-sv-research"},
    {"vcf-sv-research-index"},
]

MIP_DNA_ANALYSIS_SAMPLE_TAGS = [{"bam"}, {"bam-index"}, {"cram"}, {"cram-index"}]

MIP_RNA_ANALYSIS_CASE_TAGS = []

MIP_RNA_ANALYSIS_SAMPLE_TAGS = [
    {"bam"},
    {"bam-index"},
    {"cram"},
    {"cram-index"},
]

MICROSALT_ANALYSIS_CASE_TAGS = [
    {"microsalt-qc"},
    {"microsalt-type"},
    {"assembly"},
    {"trimmed-forward-reads"},
    {"trimmed-reverse-reads"},
    {"reference-alignment-deduplicated"},
]

MICROSALT_ANALYSIS_SAMPLE_TAGS = []

PIPELINE_ANALYSIS_TAG_MAP = {
    "balsamic": {
        "case_tags": BALSAMIC_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_ANALYSIS_SAMPLE_TAGS,
    },
    "balsamic-qc": {
        "case_tags": BALSAMIC_QC_CASE_TAGS,
        "sample_tags": BALSAMIC_QC_SAMPLE_TAGS,
    },
    "mip-dna": {
        "case_tags": MIP_DNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_DNA_ANALYSIS_SAMPLE_TAGS,
    },
    "mip-rna": {
        "case_tags": MIP_RNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_RNA_ANALYSIS_SAMPLE_TAGS,
    },
    # These are not ready yet, add when microsalt cases are stored in db
    # "microsalt": {
    #     "case_tags": MICROSALT_ANALYSIS_CASE_TAGS,
    #     "sample_tags": MICROSALT_ANALYSIS_SAMPLE_TAGS,
    # },
}

PIPELINE_ANALYSIS_OPTIONS = PIPELINE_ANALYSIS_TAG_MAP.keys()



"""Helper functions for compress cli"""
import logging
import os
from pathlib import Path
from typing import Iterator, List

from housekeeper.store import models as hk_models

from cg.apps.hk import HousekeeperAPI
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.meta.compress.files import get_spring_paths
from cg.store import Store, models
from cg.store.get.cases import ready_for_spring_compression

LOG = logging.getLogger(__name__)


def get_fastq_individuals(store: Store, case_id: str = None) -> Iterator[str]:
    """Fetch individual ids from cases that are ready for SPRING compression"""
    for case in get_fastq_cases(store, case_id):
        for link_obj in case.links:
            yield link_obj.sample.internal_id


def get_fastq_cases(store: Store, case_id: str = None):
    """Return cases ready for SPRING compression"""
    if case_id:
        return get_cases(store, case_id)
    return ready_for_spring_compression(store)


def get_cases(store: Store, case_id: str = None) -> List[models.Family]:
    """Fetch cases from store

    If case_id return one case (if existing) otherwise return all
    """
    if case_id:
        case_obj = store.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            raise CaseNotFoundError("Could not find case {}".format(case_id))
        return [case_obj]

    return store.families()


def update_compress_api(
    compress_api: CompressAPI, dry_run: bool, ntasks: int = None, mem: int = None
) -> None:
    """Update parameters in compress api"""

    compress_api.set_dry_run(dry_run)
    if ntasks:
        LOG.info("Set ntasks to %s", ntasks)
        compress_api.ntasks = ntasks
    if mem:
        LOG.info("Set mem to %s", ntasks)
        compress_api.mem = mem


# Functions to fix problematic spring files


def get_versions(hk_api: HousekeeperAPI, bundle_name: str = None) -> Iterator[hk_models.Version]:
    """Generates versions from hk bundles

    If no bundle name is given generate latest version for every bundle
    """
    if bundle_name:
        bundle = hk_api.bundle(bundle_name)
        if not bundle:
            LOG.info("Could not find bundle %s", bundle_name)
            return
        bundles = [bundle]
    else:
        bundles = hk_api.bundles()

    for bundle in bundles:
        LOG.debug("Check for versions in %s", bundle.name)
        last_version = hk_api.last_version(bundle.name)
        if not last_version:
            LOG.warning("No bundle found for %s in housekeeper", bundle.name)
            return
        yield last_version


def get_true_dir(dir_path: Path) -> Path:
    """Loop over the files in a directory, if any symlinks are found return the parent dir of the
    origin file"""
    # Check if there are any links to fastq files in the directory
    for fastq_path in dir_path.rglob("*"):
        # Check if there are fastq symlinks that points to the directory where the spring
        # path is located
        if fastq_path.is_symlink():
            true_dir = Path(os.readlink(fastq_path)).parent
            return true_dir
    LOG.info("Could not find any symlinked files")
    return None


def correct_spring_paths(
    hk_api: HousekeeperAPI, bundle_name: str = None, dry_run: bool = False
) -> None:
    """Function that will be used as a one off thing

    There has been a problem when there are symlinked fastq files that are sent for compression.
    In these cases the spring archive has been created in the same place as that the symlinks are
    pointing to. This function will find those cases and move the spring archives to the correct
    place as specified in housekeeper.
    """
    versions = get_versions(hk_api=hk_api, bundle_name=bundle_name)
    for version_obj in versions:
        spring_paths = get_spring_paths(version_obj)
        i = 0
        for i, compression_obj in enumerate(spring_paths, 1):
            # We are interested in fixing the cases where spring paths are in wrong location
            spring_path = compression_obj.spring_path
            if spring_path.exists():
                continue

            spring_config_path = compression_obj.spring_metadata_path
            # true_dir is where the spring paths actually exists
            true_dir = get_true_dir(spring_path.parent)
            if not true_dir:
                LOG.info("Could not find location of spring files")
                continue

            true_spring_path = true_dir / spring_path.name
            true_spring_config_path = true_spring_path.with_suffix("").with_suffix(".json")
            if not (true_spring_path.exists() and true_spring_config_path.exists()):
                LOG.info("Could not find spring and/or spring metadata files, skipping")
                continue
            LOG.info(
                "Moving existing spring file (and config) %s to hk bundle path %s",
                true_spring_path,
                spring_path,
            )
            if not dry_run:
                # We know from above that the spring path does not exist
                true_spring_path.replace(spring_path)
                true_spring_config_path.replace(spring_config_path)
        if i == 0:
            LOG.debug("Could not find any spring files")



"""Fixtures for meta/upload tests"""

import json

from pathlib import Path
from datetime import datetime

import pytest

from cg.apps.coverage.api import ChanjoAPI
from cg.constants import Pipeline
from cg.meta.upload.coverage import UploadCoverageApi
from cg.meta.upload.mutacc import UploadToMutaccAPI
from cg.meta.upload.observations import UploadObservationsAPI
from cg.meta.upload.scoutapi import UploadScoutAPI
from cg.meta.upload.gt import UploadGenotypesAPI
from cg.store import models
from cg.store import Store


# Mocks


class MockAnalysis:
    """Mock an analysis object"""

    @staticmethod
    def get_latest_metadata(family_id=None):
        """Mock get_latest_metadata"""
        # Returns: dict: parsed data
        # Define output dict
        outdata = {
            "analysis_sex": {"ADM1": "female", "ADM2": "female", "ADM3": "female"},
            "family": family_id or "yellowhog",
            "duplicates": {"ADM1": 13.525, "ADM2": 12.525, "ADM3": 14.525},
            "genome_build": "hg19",
            "rank_model_version": "1.18",
            "mapped_reads": {"ADM1": 98.8, "ADM2": 99.8, "ADM3": 97.8},
            "mip_version": "v4.0.20",
            "sample_ids": ["2018-20203", "2018-20204"],
            "sv_rank_model_version": "1.08",
        }
        return outdata

    @staticmethod
    def convert_panels(customer_id, panels):
        """Mock convert_panels"""
        _ = customer_id, panels
        return ""


class MockCoverage(ChanjoAPI):
    """Mock chanjo coverage api"""


class MockMutaccAuto:
    """ Mock class for mutacc_auto api """

    @staticmethod
    def extract_reads(*args, **kwargs):
        """mock extract_reads method"""
        _, _ = args, kwargs

    @staticmethod
    def import_reads(*args, **kwargs):
        """ mock import_reads method """
        _, _ = args, kwargs


class MockScoutApi:
    """Mock class for Scout api"""

    @staticmethod
    def get_causative_variants(case_id):
        """mock get_causative_variants"""
        _ = case_id
        return []

    @property
    def client(self):
        """Client URI"""
        return "mongodb://"

    @property
    def name(self):
        """Name property"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @name.deleter
    def name(self):
        del self._name


class MockLoqusAPI:
    """ Mock LoqusAPI class"""

    def __init__(self, analysis_type="wgs"):
        self.analysis_type = analysis_type

    @staticmethod
    def load(*args, **kwargs):
        """ Mock load method"""
        _ = args
        _ = kwargs
        return dict(variants=12)

    @staticmethod
    def get_case(*args, **kwargs):
        """ Mock get_case method"""
        _ = args
        _ = kwargs
        return {"case_id": "case_id", "_id": "123"}

    @staticmethod
    def get_duplicate(*args, **kwargs):
        """Mock get_duplicate method"""
        _ = args
        _ = kwargs
        return {"case_id": "case_id"}


class MockLims:
    """ Mock Lims API """

    lims = None

    def __init__(self, samples):
        self.lims = self
        self._samples = samples

    def sample(self, sample_id):
        """ Returns a lims sample matching the provided sample_id """
        for sample in self._samples:
            if sample["id"] == sample_id:
                return sample
        return None


@pytest.fixture(name="lims_family")
def fixture_lims_family():
    """ Returns a lims-like family of samples """
    return json.load(open("tests/fixtures/report/lims_family.json"))


@pytest.fixture(name="lims_samples")
def fixture_lims_samples(lims_family):
    """ Returns the samples of a lims family """
    return lims_family["samples"]


@pytest.fixture(name="upload_genotypes_hk_bundle")
def fixture_upload_genotypes_hk_bundle(
    case_id: str, timestamp, case_qc_metrics: Path, bcf_file: Path
) -> dict:
    """ Returns a dictionary in hk format with files used in upload gt process"""
    data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {"path": str(case_qc_metrics), "archive": False, "tags": ["qcmetrics"]},
            {"path": str(bcf_file), "archive": False, "tags": ["snv-gbcf"]},
        ],
    }
    return data


@pytest.fixture(name="analysis_obj")
def fixture_analysis_obj(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers
) -> models.Analysis:
    """Return a analysis object with a trio"""
    family_obj = analysis_store_trio.family(case_id)
    helpers.add_analysis(store=analysis_store_trio, family=family_obj, started_at=timestamp)
    return analysis_store_trio.family(case_id).analyses[0]


@pytest.yield_fixture(name="upload_genotypes_api")
def fixture_upload_genotypes_api(
    real_housekeeper_api, genotype_api, upload_genotypes_hk_bundle, helpers
) -> UploadGenotypesAPI:
    """Create a upload genotypes api"""
    helpers.ensure_hk_bundle(real_housekeeper_api, upload_genotypes_hk_bundle, include=True)
    _api = UploadGenotypesAPI(
        hk_api=real_housekeeper_api,
        gt_api=genotype_api,
    )

    return _api


@pytest.yield_fixture(scope="function")
def upload_observations_api(analysis_store, populated_housekeeper_api):
    """ Create mocked UploadObservationsAPI object"""

    loqus_mock = MockLoqusAPI()

    _api = UploadObservationsAPI(
        status_api=analysis_store,
        hk_api=populated_housekeeper_api,
        loqus_api=loqus_mock,
    )

    yield _api


@pytest.yield_fixture(scope="function")
def upload_observations_api_wes(analysis_store, populated_housekeeper_api):
    """ Create mocked UploadObservationsAPI object"""

    loqus_mock = MockLoqusAPI(analysis_type="wes")

    _api = UploadObservationsAPI(
        status_api=analysis_store,
        hk_api=populated_housekeeper_api,
        loqus_api=loqus_mock,
    )

    yield _api


@pytest.yield_fixture(scope="function")
def upload_scout_api(scout_api, madeline_api, lims_samples, populated_housekeeper_api):
    """Fixture for upload_scout_api"""
    analysis_mock = MockAnalysis()
    lims_api = MockLims(lims_samples)

    _api = UploadScoutAPI(
        hk_api=populated_housekeeper_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
    )

    yield _api


@pytest.yield_fixture(scope="function")
def mutacc_upload_api():
    """
    Fixture for a mutacc upload api
    """

    scout_api = MockScoutApi()
    mutacc_auto_api = MockMutaccAuto()

    _api = UploadToMutaccAPI(scout_api=scout_api, mutacc_auto_api=mutacc_auto_api)

    return _api


@pytest.yield_fixture(scope="function")
def coverage_upload_api(chanjo_config_dict, populated_housekeeper_api):
    """Fixture for coverage upload API"""
    hk_api = populated_housekeeper_api
    status_api = None
    coverage_api = MockCoverage(chanjo_config_dict)
    _api = UploadCoverageApi(status_api=status_api, hk_api=hk_api, chanjo_api=coverage_api)
    return _api


@pytest.yield_fixture(scope="function")
def analysis(analysis_store, case_id):
    """Fixture to mock an analysis"""
    _analysis = analysis_store.add_analysis(pipeline=Pipeline.BALSAMIC, version="version")
    _analysis.family = analysis_store.family(case_id)
    _analysis.config_path = "dummy_path"
    yield _analysis



"""
This module handles concatenation of usalt fastq files.

Classes:
    FastqFileNameCreator: Creates valid usalt filenames
    FastqHandler: Handles fastq file linking
"""
import logging
from typing import List

LOGGER = logging.getLogger(__name__)


class BaseFastqHandler:
    """Handles fastq file linking"""

    class BaseFastqFileNameCreator:
        """Creates valid usalt filename from the parameters"""

        @staticmethod
        def create(lane: str, flowcell: str, sample: str, read: str, more: dict = None) -> str:
            """Name a FASTQ file following pipeline conventions."""

    def __init__(self, config):
        pass

    def link(self, case: str, sample: str, files: List[dict]):
        """Link FASTQ files for a pipeline sample."""



"""API to run Vogue"""

import json

from cg.apps.gt import GenotypeAPI
from cg.apps.vogue import VogueAPI
from cg.store import Store


class UploadVogueAPI:
    """API to load data into Vogue"""

    def __init__(self, genotype_api: GenotypeAPI, vogue_api: VogueAPI, store: Store):
        self.genotype_api = genotype_api
        self.vogue_api = vogue_api
        self.store = store

    def load_genotype(self, days: int) -> None:
        """Loading genotype data from the genotype database into the trending database"""
        samples = self.genotype_api.export_sample(days=days)
        samples = json.loads(samples)
        for sample_id, sample_dict in samples.items():
            sample_dict["_id"] = sample_id
            self.vogue_api.load_genotype_data(sample_dict)

        samples_analysis = self.genotype_api.export_sample_analysis(days=days)
        samples_analysis = json.loads(samples_analysis)
        for sample_id, sample_dict in samples_analysis.items():
            sample_dict["_id"] = sample_id
            self.vogue_api.load_genotype_data(sample_dict)

    def load_apptags(self) -> None:
        """Loading application tags from statusdb into the trending database"""
        apptags = self.store.applications()
        apptags_for_vogue = []
        for tag in apptags.all():
            apptags_for_vogue.append({"tag": tag.tag, "prep_category": tag.prep_category})

        self.vogue_api.load_apptags(apptags_for_vogue)

    def load_bioinfo_raw(self, load_bioinfo_inputs):
        """Running vogue load bioinfo raw."""

        self.vogue_api.load_bioinfo_raw(load_bioinfo_inputs)

    def load_bioinfo_process(self, load_bioinfo_inputs, cleanup):
        """Running vogue load bioinfo process."""

        self.vogue_api.load_bioinfo_process(load_bioinfo_inputs, cleanup)

    def load_bioinfo_sample(self, load_bioinfo_inputs):
        """Running vogue load bioinfo sample."""

        self.vogue_api.load_bioinfo_sample(load_bioinfo_inputs)



from cg.apps.lims import LimsAPI
from cg.server.ext import lims as genologics_lims
from cg.store import Store, models


class InvoiceAPI:
    def __init__(self, db: Store, lims_api: LimsAPI, invoice_obj: models.Invoice):
        self.db = db
        self.lims_api = lims_api
        self.log = []
        self.invoice_obj = invoice_obj
        self.customer_obj = invoice_obj.customer
        self.record_type = ""
        self.raw_records = []
        self._set_record_type()

    def _set_record_type(self):
        """Define the record_type based on the invoice object.
        It can only be either pool or sample"""

        if self.invoice_obj.pools:
            self.raw_records = self.invoice_obj.pools
            self.record_type = "Pool"
        elif self.invoice_obj.samples:
            self.record_type = "Sample"
            self.raw_records = self.invoice_obj.samples

    def prepare_contact_info(self, costcenter):
        """Function to prepare contact info for a customer"""

        msg = (
            f"Could not open/generate invoice. Contact information missing in database for "
            f"customer {self.customer_obj.internal_id}. See log files."
        )

        customer = self.db.customer("cust999") if costcenter.lower() == "kth" else self.customer_obj
        user = customer.invoice_contact

        if not user:
            self.log.append(msg)
            return None

        contact = {
            "name": user.name,
            "email": user.email,
            "customer_name": customer.name,
            "reference": customer.invoice_reference,
            "address": customer.invoice_address,
        }

        if None in contact.values():
            self.log.append(msg)
            return None

        return contact

    def prepare(self, costcenter: str) -> dict:
        """Get information about an invoice to generate Excel report."""

        records = []
        pooled_samples = []

        for raw_record in self.raw_records:
            if self.record_type == "Pool":
                pooled_samples += genologics_lims.samples_in_pools(
                    raw_record.name, raw_record.ticket_number
                )
            record = self.prepare_record(
                costcenter=costcenter.lower(), discount=self.invoice_obj.discount, record=raw_record
            )
            if record:
                records.append(record)
            else:
                return None

        customer_obj = self.invoice_obj.customer
        contact = self.prepare_contact_info(costcenter)
        if not contact:
            return None
        return {
            "costcenter": costcenter,
            "project_number": getattr(customer_obj, f"project_account_{costcenter.lower()}"),
            "customer_id": customer_obj.internal_id,
            "customer_name": customer_obj.name,
            "agreement": customer_obj.agreement_registration,
            "invoice_id": self.invoice_obj.id,
            "contact": contact,
            "records": records,
            "pooled_samples": pooled_samples,
            "record_type": self.record_type,
        }

    def _discount_price(self, record, discount: int = 0):
        """Get discount price for a sample or pool."""

        if self.record_type == "Pool" or record.priority_human == "clinical trials":
            priority = "research"
        else:
            priority = record.priority_human

        full_price = getattr(record.application_version, f"price_{priority}")
        discount_factor = float(100 - discount) / 100

        if not full_price:
            return None
        return full_price * discount_factor

    def _cost_center_split_factor(self, price, costcenter, percent_kth, tag, version):
        """Split price based on cost center"""
        if price:
            try:
                if costcenter == "kth":
                    split_factor = percent_kth / 100
                else:
                    split_factor = (100 - percent_kth) / 100
                split_price = round(price * split_factor, 1)
            except ValueError:
                self.log.append(
                    f"Could not calculate price for samples with application "
                    f"tag/version: {tag}/{version}. Missing %KTH"
                )
                return None
        else:
            self.log.append(
                f"Could not get price for samples with application tag/version: {tag}/{version}."
            )
            return None
        return split_price

    def prepare_record(self, costcenter: str, discount: int, record):
        """Get invoice information for a specific sample or pool"""

        try:
            tag = record.application_version.application.tag
            version = str(record.application_version.version)
            percent_kth = record.application_version.application.percent_kth
            discounted_price = self._discount_price(record, discount)
        except ValueError:
            self.log.append(f"Application tag/version seems to be missing for sample {record.id}.")
            return None

        split_discounted_price = self._cost_center_split_factor(
            discounted_price, costcenter, percent_kth, tag, version
        )

        order = record.order
        ticket_number = record.ticket_number
        lims_id = None if self.record_type == "Pool" else record.internal_id
        priority = "research" if self.record_type == "Pool" else record.priority_human

        invoice_info = {
            "name": record.name,
            "lims_id": lims_id,
            "id": record.id,
            "application_tag": record.application_version.application.tag,
            "project": f"{order or 'NA'} ({ticket_number or 'NA'})",
            "date": record.received_at.date() if record.received_at else "",
            "price": split_discounted_price,
            "priority": priority,
        }
        if costcenter == "ki":
            price_kth = self._cost_center_split_factor(
                discounted_price, "kth", percent_kth, tag, version
            )
            invoice_info["price_kth"] = price_kth
            invoice_info["total_price"] = discounted_price
        return invoice_info

    def total_price(self) -> float:
        """Get the total price for all records in the invoice"""

        discount = self.invoice_obj.discount
        total_price = 0

        for record in self.raw_records:
            discount_price = self._discount_price(record, discount)
            if discount_price:
                total_price += discount_price
            else:
                return None

        return total_price



"""Code that handles uploading to mutacc from the CLI"""
import logging

import click

from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.meta.upload.mutacc import UploadToMutaccAPI

LOG = logging.getLogger(__name__)


@click.command("process-solved")
@click.option("-c", "--case-id", help="internal case id, leave empty to process all")
@click.option("-d", "--days-ago", type=int, default=1, help="days since solved")
@click.option("-C", "--customers", type=str, multiple=True, help="Filter on customers")
@click.option("--dry-run", is_flag=True, help="only print cases to be processed")
@click.pass_context
def process_solved(context, case_id, days_ago, customers, dry_run):
    """Process cases with mutacc that has been marked as solved in scout.
    This prepares them to be uploaded to the mutacc database"""

    click.echo(click.style("----------------- PROCESS-SOLVED ----------------"))

    scout_api = context.obj["scout_api"]
    mutacc_auto_api = MutaccAutoAPI(context.obj)

    mutacc_upload = UploadToMutaccAPI(scout_api=scout_api, mutacc_auto_api=mutacc_auto_api)

    # Get cases to upload into mutacc from scout
    if case_id is not None:
        finished_cases = scout_api.get_cases(finished=True, case_id=case_id)
    elif days_ago is not None:
        finished_cases = scout_api.get_solved_cases(days_ago=days_ago)
    else:
        LOG.info("Please enter option '--case-id' or '--days-ago'")

    number_processed = 0
    for case in finished_cases:

        number_processed += 1
        if customers:
            if case["owner"] not in customers:
                LOG.info("skipping %s: Not valid customer %s", case["_id"], case["owner"])
                continue
        if dry_run:
            LOG.info("Would process case %s with mutacc", case["_id"])
            continue

        LOG.info("Start processing case %s with mutacc", case["_id"])
        mutacc_upload.extract_reads(case)

    if number_processed == 0:
        LOG.info("No cases were solved within the last %s days", days_ago)


@click.command("processed-solved")
@click.pass_context
def processed_solved(context):
    """Upload solved cases that has been processed by mutacc to the mutacc database"""

    click.echo(click.style("----------------- PROCESSED-SOLVED ----------------"))

    LOG.info("Uploading processed cases by mutacc to the mutacc database")

    mutacc_auto_api = MutaccAutoAPI(context.obj)
    mutacc_auto_api.import_reads()



"""Code for talking to Scout regarding uploads"""

import datetime as dt
import logging
from typing import List
from pathlib import Path

from pymongo import MongoClient
from scout.adapter.mongo import MongoAdapter
from scout.export.panel import export_panels as scout_export_panels
from scout.load import load_scout
from scout.load.report import load_delivery_report
from scout.parse.case import parse_case_data
from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


class ScoutAPI(MongoAdapter):

    """Interface to Scout."""

    def __init__(self, config):
        client = MongoClient(config["scout"]["database"], serverSelectionTimeoutMS=20)
        super(ScoutAPI, self).__init__(client[config["scout"]["database_name"]])

        binary_path = config["scout"]["binary_path"]
        config_path = config["scout"]["config_path"]
        self.process = Process(binary=binary_path, config=config_path)

    def upload(self, data: dict, threshold: int = 5, force: bool = False):
        """Load analysis of a new family into Scout."""
        data["rank_score_threshold"] = threshold
        config_data = parse_case_data(config=data)
        existing_case = self.case(
            institute_id=config_data["owner"], display_name=config_data["family_name"]
        )
        if existing_case:
            if force or config_data["analysis_date"] > existing_case["analysis_date"]:
                LOG.info("update existing Scout case")
                load_scout(self, config_data, update=True)
            else:
                existing_date = existing_case["analysis_date"].date()
                LOG.warning("analysis of case already loaded: %s", existing_date)
            return
        LOG.debug("load new Scout case")
        load_scout(self, config_data)
        LOG.debug("Case loaded successfully to Scout")

    def update_alignment_file(self, case_id: str, sample_id: str, alignment_path: Path):
        """Update alignment file for individual in case"""
        parameters = [
            "update",
            "individual",
            "--case-id",
            case_id,
            "--ind-id",
            sample_id,
            "--alignment-path",
            str(alignment_path),
        ]
        self.process.run_command(parameters=parameters)

    def export_panels(self, panels: List[str], versions=None):
        """Pass through to export of a list of gene panels."""
        return scout_export_panels(self, panels, versions)

    def get_genes(self, panel_id: str, version: str = None) -> list:
        """Fetch panel genes.

        Args:
            panel_id (str): unique id for the panel
            version (str): version of the panel. If 'None' latest version will be returned

        Returns:
            panel genes: panel genes list
        """
        gene_panel = self.gene_panel(panel_id=panel_id, version=version)
        return gene_panel.get("genes")

    def get_cases(
        self,
        case_id=None,
        institute=None,
        reruns=None,
        finished=None,
        causatives=None,
        research_requested=None,
        is_research=None,
        status=None,
    ):
        """Interact with cases existing in the database."""

        models = []
        if case_id:
            case_obj = self.case(case_id=case_id)
            if case_obj:
                models.append(case_obj)

        else:
            models = self.cases(
                collaborator=institute,
                reruns=reruns,
                finished=finished,
                has_causatives=causatives,
                research_requested=research_requested,
                is_research=is_research,
                status=status,
            )

        return models

    def get_causative_variants(self, case_id=None, collaborator=None):
        """
        Get causative variants for a case
        """
        causative_ids = self.get_causatives(institute_id=collaborator, case_id=case_id)

        causatives = [self.variant(causative_id) for causative_id in causative_ids]

        return causatives

    def get_solved_cases(self, days_ago):
        """
        Get cases solved within chosen timespan

        Args:
            days_ago (int): Maximum days ago a case has been solved

        Return:
            cases (list): list of cases
        """

        days_datetime = dt.datetime.now() - dt.timedelta(days=days_ago)

        # Look up 'mark_causative' events added since specified number days ago
        event_query = {
            "category": "case",
            "verb": "mark_causative",
            "created_at": {"$gte": days_datetime},
        }
        recent_events = self.event_collection.find(event_query)
        solved_cases = set()

        # Find what cases these events concern
        for event in recent_events:
            solved_cases.add(event["case"])

        solved_cases = list(solved_cases)

        # Find these cases in the database
        cases = self.case_collection.find({"_id": {"$in": solved_cases}})

        return cases

    def upload_delivery_report(self, report_path: str, case_id: str, update: bool = False):
        """Load a delivery report into a case in the database

        If the report already exists the function will exit.
        If the user want to load a report that is already in the database
        'update' has to be 'True'

        Args:
            report_path (string):       Path to delivery report
            case_id     (string):       Case identifier
            update      (bool):         If an existing report should be replaced

        Returns:
            updated_case(dict)

        """

        return load_delivery_report(
            adapter=self, case_id=case_id, report_path=report_path, update=update
        )



"""Click commands to store mip analyses"""
import logging
from pathlib import Path

import click

from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.exc import (
    AnalysisNotFinishedError,
    AnalysisDuplicationError,
    BundleAlreadyAddedError,
    PipelineUnknownError,
    MandatoryFilesMissing,
)
from cg.meta.store.base import gather_files_and_bundle_in_housekeeper
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.store import Store
from cg.constants import EXIT_SUCCESS, EXIT_FAIL, Pipeline


LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def store(context):
    """Store results from MIP in housekeeper."""
    context.obj["housekeeper_api"] = HousekeeperAPI(context.obj)
    context.obj["trailblazer_api"] = TrailblazerAPI(context.obj)
    context.obj["scout_api"] = ScoutAPI(context.obj)
    context.obj["lims_api"] = LimsAPI(context.obj)
    context.obj["status_db"] = Store(context.obj["database"])

    context.obj["mip_api"] = MipAnalysisAPI(
        db=context.obj["status_db"],
        hk_api=context.obj["housekeeper_api"],
        tb_api=context.obj["trailblazer_api"],
        scout_api=context.obj["scout_api"],
        lims_api=context.obj["lims_api"],
        script=context.obj["mip-rd-dna"]["script"],
        pipeline=context.obj["mip-rd-dna"]["pipeline"],
        conda_env=context.obj["mip-rd-dna"]["conda_env"],
        root=context.obj["mip-rd-dna"]["root"],
    )


@store.command()
@click.argument("config-stream", type=click.File("r"), required=False)
@click.pass_context
def analysis(context, config_stream):
    """Store a finished analysis in Housekeeper."""
    mip_api = context.obj["mip_api"]

    exit_code = EXIT_SUCCESS
    if not config_stream:
        LOG.error("Provide a config file.")
        raise click.Abort

    try:
        new_analysis = gather_files_and_bundle_in_housekeeper(
            config_stream,
            mip_api.hk,
            mip_api.db,
            workflow=Pipeline.MIP_DNA,
        )
        mip_api.db.add_commit(new_analysis)
    except (
        AnalysisNotFinishedError,
        AnalysisDuplicationError,
        BundleAlreadyAddedError,
        PipelineUnknownError,
        MandatoryFilesMissing,
    ) as error:
        LOG.error(error.message)
        exit_code = EXIT_FAIL
    except FileNotFoundError as error:
        LOG.error(f"Missing file: {error.args[0]}")
        exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort

    LOG.info("Included files in Housekeeper")


@store.command()
@click.pass_context
def completed(context):
    """Store all completed analyses."""
    mip_api = context.obj["mip_api"]

    exit_code = EXIT_SUCCESS
    for case_obj in mip_api.db.cases_to_store(pipeline=Pipeline.MIP_DNA):
        try:
            analysis_obj = mip_api.tb.get_latest_analysis(case_id=case_obj.internal_id)
            if analysis_obj.status != "completed":
                continue
            LOG.info(f"Storing case: {analysis_obj.family}")
            with Path(
                mip_api.get_case_config_path(case_id=analysis_obj.family)
            ).open() as config_stream:

                context.invoke(analysis, config_stream=config_stream)
        except (Exception, click.Abort):
            LOG.error(f"Case storage failed: {case_obj.internal_id}", exc_info=True)
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort



""" API to manage Microsalt Analyses
    Organism - Fallback based on reference, ‘Other species’ and ‘Comment’. Default to “Unset”.
    Priority = Default to empty string. Weird response. Typically “standard” or “research”.
    Reference = Defaults to “None”
    Method: Outputted as “1273:23”. Defaults to “Not in LIMS”
    Date: Returns latest == most recent date. Outputted as DT object “YYYY MM DD”. Defaults to
    datetime.min"""
import gzip
import logging
from pathlib import Path
import re
from datetime import datetime
from typing import Dict, List

from cg.apps.microsalt.fastq import FastqHandler
from cg.constants import FAMILY_ACTIONS
from cg.exc import CgDataError
from cg.store.models import Sample

from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.store import models, Store

LOG = logging.getLogger(__name__)


class MicrosaltAnalysisAPI:
    """API to manage Microsalt Analyses"""

    def __init__(
        self,
        db: Store,
        hk_api: HousekeeperAPI,
        lims_api: LimsAPI,
        fastq_handler: FastqHandler,
        config: dict = None,
    ):
        self.db = db
        self.hk = hk_api
        self.lims = lims_api
        self.fastq_handler = fastq_handler
        self.root_dir = config.get("root") if config else None

    def has_flowcells_on_disk(self, ticket: int) -> bool:
        """Check stuff before starting the analysis."""

        flowcells = self.get_flowcells(ticket=ticket)
        statuses = []
        for flowcell_obj in flowcells:
            LOG.debug(f"{flowcell_obj.name}: checking if flowcell is on disk")
            statuses.append(flowcell_obj.status)
            if flowcell_obj.status == "removed":
                LOG.info(f"{flowcell_obj.name}: flowcell not on disk")
            elif flowcell_obj.status != "ondisk":
                LOG.warning(f"{flowcell_obj.name}: {flowcell_obj.status}")
        return all(status == "ondisk" for status in statuses)

    @staticmethod
    def fastq_header(line: str) -> dict:
        """handle illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence
        """

        rs = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            rs["lane"] = parts[1]
            rs["flowcell"] = "XXXXXX"
            rs["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[-1].split("/")[-1]

        return rs

    def link_samples(self, ticket: int, sample_id: str):

        sample_objs = self.get_samples(ticket, sample_id)

        if not sample_objs:
            raise Exception("Could not find any samples to link")

        for sample_obj in sample_objs:
            LOG.info("%s: link FASTQ files", sample_obj.internal_id)
            self.link_sample(
                ticket=ticket,
                sample=sample_obj.internal_id,
            )

    def link_sample(self, sample: str, ticket: int) -> None:
        """Link FASTQ files for a sample."""
        file_objs = self.hk.files(bundle=sample, tags=["fastq"])
        files = []

        for file_obj in file_objs:
            # figure out flowcell name from header
            with gzip.open(file_obj.full_path) as handle:
                header_line = handle.readline().decode()
                header_info = self.fastq_header(header_line)

            data = {
                "path": file_obj.full_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("_Undetermined_" in file_obj.path),
            }
            # look for tile identifier (HiSeq X runs)
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", file_obj.path)
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            files.append(data)

        self.fastq_handler.link(ticket=ticket, sample=sample, files=files)

    def get_samples(self, ticket: int = None, sample_id: str = None) -> [models.Sample]:
        ids = {}
        if ticket:
            ids["ticket_number"] = ticket
            samples = self.db.samples_by_ids(**ids).all()
        if sample_id:
            sample = self.db.sample(internal_id=sample_id)
            samples = [sample] if sample else []
        return samples

    def get_lims_comment(self, sample_id: str) -> str:
        """ returns the comment associated with a sample stored in lims"""
        comment = self.lims.get_sample_comment(sample_id) or ""
        if re.match(r"\w{4}\d{2,3}", comment):
            return comment

        return ""

    def get_organism(self, sample_obj: models.Sample) -> str:
        """Organism
        - Fallback based on reference, ‘Other species’ and ‘Comment’.
        Default to "Unset"."""

        if not sample_obj.organism:
            raise CgDataError(f"Organism missing on Sample")

        organism = sample_obj.organism.internal_id.strip()
        comment = self.get_lims_comment(sample_id=sample_obj.internal_id)
        has_comment = bool(comment)

        if "gonorrhoeae" in organism:
            organism = "Neisseria spp."
        elif "Cutibacterium acnes" in organism:
            organism = "Propionibacterium acnes"

        if organism == "VRE":
            reference = sample_obj.organism.reference_genome
            if reference == "NC_017960.1":
                organism = "Enterococcus faecium"
            elif reference == "NC_004668.1":
                organism = "Enterococcus faecalis"
            elif has_comment:
                organism = comment

        return organism

    def get_parameters(self, sample_obj: Sample) -> Dict[str, str]:
        """Fill a dict with case config information for one sample """

        sample_id = sample_obj.internal_id
        method_library_prep = self.lims.get_prep_method(sample_id)
        if method_library_prep:
            method_library_prep, _ = method_library_prep.split(" ", 1)
        method_sequencing = self.lims.get_sequencing_method(sample_id)
        if method_sequencing:
            method_sequencing, _ = method_sequencing.split(" ", 1)
        priority = "research" if sample_obj.priority == 0 else "standard"

        parameter_dict = {
            "CG_ID_project": self.get_project(sample_id),
            "Customer_ID_project": sample_obj.ticket_number,
            "CG_ID_sample": sample_obj.internal_id,
            "Customer_ID_sample": sample_obj.name,
            "organism": self.get_organism(sample_obj),
            "priority": priority,
            "reference": sample_obj.reference_genome,
            "Customer_ID": sample_obj.customer.internal_id,
            "application_tag": sample_obj.application_version.application.tag,
            "date_arrival": str(sample_obj.received_at or datetime.min),
            "date_sequencing": str(sample_obj.sequenced_at or datetime.min),
            "date_libprep": str(sample_obj.prepared_at or datetime.min),
            "method_libprep": method_library_prep or "Not in LIMS",
            "method_sequencing": method_sequencing or "Not in LIMS",
        }

        return parameter_dict

    def get_project(self, sample_id: str) -> str:
        """Get LIMS project for a sample"""
        return self.lims.get_sample_project(sample_id)

    def get_flowcells(self, ticket: int) -> [models.Flowcell]:
        """Get all flowcells for all samples in a ticket"""

        flowcells = set()

        for sample in self.get_samples(ticket=ticket):
            for flowcell in sample.flowcells:
                flowcells.add(flowcell)

        return list(flowcells)

    def request_removed_flowcells(self, ticket):
        """Check stuff before starting the analysis."""

        flowcells = self.get_flowcells(ticket=ticket)
        for flowcell_obj in flowcells:
            LOG.debug(f"{flowcell_obj.name}: checking if flowcell should be requested")
            if flowcell_obj.status == "removed":
                LOG.info(f"{flowcell_obj.name}: requesting removed flowcell")
                flowcell_obj.status = "requested"
            elif flowcell_obj.status != "ondisk":
                LOG.warning(f"{flowcell_obj.name}: {flowcell_obj.status}")
        self.db.commit()

    def get_deliverables_to_store(self) -> List[Path]:
        """Retrieve a list of microbial deliverables files for orders where analysis finished
        successfully, and are ready to be stored in Housekeeper"""
        deliverables_to_store = []
        for case_object in self.db.cases_to_store(pipeline="microbial"):
            deliverables_file = self.get_deliverables_file_path(order_id=case_object.name)
            if not deliverables_file.exists():
                continue
            deliverables_to_store.append(deliverables_file)
        return deliverables_to_store

    def get_deliverables_file_path(self, order_id: str) -> Path:
        """Returns a path where the microSALT deliverables file for the order_id should be
        located"""
        deliverables_file_path = Path(
            self.root_dir,
            "results/reports/deliverables",
            order_id + "_deliverables.yaml",
        )
        return deliverables_file_path

    def set_statusdb_action(self, name: str, action: str) -> None:
        """Sets action on case based on ticket number"""
        if action in [None, *FAMILY_ACTIONS]:
            case_object = self.db.find_family(name)
            case_object.action = action
            self.db.commit()
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {name}"
        )



"""Module for Balsamic Analysis API"""

import datetime as dt
import gzip
import json
import logging
import re
from pathlib import Path
from typing import List, Optional

from cg.apps.balsamic.api import BalsamicAPI
from cg.apps.balsamic.fastq import FastqHandler

from cg.constants import FAMILY_ACTIONS, Pipeline
from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.exc import BalsamicStartError, BundleAlreadyAddedError
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class BalsamicAnalysisAPI:
    """Handles communication between BALSAMIC processes
    and the rest of CG infrastructure"""

    __BALSAMIC_APPLICATIONS = {"wgs", "wes", "tgs"}
    __BALSAMIC_BED_APPLICATIONS = {"wes", "tgs"}

    def __init__(
        self,
        balsamic_api: BalsamicAPI,
        store: Store,
        housekeeper_api: HousekeeperAPI,
        fastq_handler: FastqHandler,
        lims_api: LimsAPI,
        trailblazer_api: TrailblazerAPI,
    ):
        self.balsamic_api = balsamic_api
        self.store = store
        self.housekeeper_api = housekeeper_api
        self.fastq_handler = fastq_handler
        self.lims_api = lims_api
        self.trailblazer_api = trailblazer_api

    def get_case_object(self, case_id: str):
        """Look up case ID in StoreDB and return result"""
        case_object = self.store.family(case_id)
        if not case_object:
            raise BalsamicStartError(f"{case_id} not found in StatusDB!")
        if not case_object.links:
            raise BalsamicStartError(
                f"{case_id} number of samples is {len(case_object.links)}, analysis will not be started!"
            )
        return case_object

    def set_statusdb_action(self, case_id: str, action: str) -> None:
        if action in [None, *FAMILY_ACTIONS]:
            case_object = self.get_case_object(case_id=case_id)
            case_object.action = action
            self.store.commit()

    def get_case_path(self, case_id: str) -> str:
        """Returns a path where the Balsamic case for the case_id should be located"""
        return Path(self.balsamic_api.root_dir, case_id).as_posix()

    def get_deliverables_file_path(self, case_id: str, check_exists: bool = False) -> str:
        """Returns a path where the Balsamic deliverables file for the case_id should be
        located.
        (Optional) Checks if deliverables file exists"""
        deliverables_file_path = Path(
            self.balsamic_api.root_dir,
            case_id,
            "analysis",
            "delivery_report",
            case_id + ".hk",
        )
        if check_exists and not deliverables_file_path.exists():
            raise BalsamicStartError(
                f"No deliverables file found for {case_id}."
                f" Make sure the deliverables file is generated, and try again!"
            )
        return deliverables_file_path.as_posix()

    def get_config_path(self, case_id: str, check_exists: bool = False) -> str:
        """Generates a path where the Balsamic config for the case_id should be located.
        (Optional) Checks if config file exists."""
        config_path = Path(self.balsamic_api.root_dir, case_id, case_id + ".json")
        if check_exists and not config_path.exists():
            raise BalsamicStartError(
                f"No config file found for {case_id}. Make sure the config file is generated, and try again!"
            )
        return config_path.as_posix()

    def get_analysis_finish_path(self, case_id: str, check_exists: bool = False) -> str:
        """Returns path to analysis_finish file.
        (Optional) Checks if analysis_finish file exists"""
        analysis_finish_path = Path(
            self.balsamic_api.root_dir, case_id, "analysis", "analysis_finish"
        )
        if check_exists and not analysis_finish_path.exists():
            raise BalsamicStartError(
                f"Analysis incomplete for {case_id}, deliverables file will not be created. "
                f"Please ensure all jobs have finished successfully!"
            )
        return analysis_finish_path.as_posix()

    def get_slurm_job_ids_path(self, case_id: str) -> Path:
        slurm_job_ids_path = Path(
            self.balsamic_api.root_dir, case_id, "analysis", "slurm_jobids.yaml"
        )
        return slurm_job_ids_path

    def get_file_collection(self, sample_id: str) -> list:
        """Retrieves sample data for naming"""
        file_objs = self.housekeeper_api.files(bundle=sample_id, tags=["fastq"])
        files = []
        for file_obj in file_objs:
            with gzip.open(file_obj.full_path) as handle:
                header_line = handle.readline().decode()
                header_info = self.fastq_handler.parse_header(header_line)

            data = {
                "path": file_obj.full_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("_Undetermined_" in file_obj.path),
            }
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", file_obj.path)
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            files.append(data)
        return files

    def get_balsamic_sample_objects(self, case_id: str) -> List[models.FamilySample]:
        """Retrieves all links where analysis is set to Balsamic"""
        case_object = self.get_case_object(case_id=case_id)
        valid_sample_list = []
        for link in case_object.links:
            if str(Pipeline.BALSAMIC) == case_object.data_analysis:
                valid_sample_list.append(link)
        return valid_sample_list

    def link_samples(self, case_id: str) -> None:
        """Links and copies files to working directory"""
        for link_object in self.get_balsamic_sample_objects(case_id=case_id):
            LOG.info(
                f"{link_object.sample.internal_id}: {link_object.family.data_analysis} linking FASTQ files"
            )
            file_collection = self.get_file_collection(sample_id=link_object.sample.internal_id)
            self.fastq_handler.link(
                case=link_object.family.internal_id,
                sample=link_object.sample.internal_id,
                files=file_collection,
            )
        LOG.info("Linking completed")

    def get_target_bed_from_lims(self, link_object: models.FamilySample) -> str:
        """Get target bed filename from lims
        Raises LimsDataError if target_bed cannot be retrieved.
        """
        capture_kit = self.lims_api.capture_kit(link_object.sample.internal_id)
        if capture_kit:
            panel_bed = self.store.bed_version(capture_kit).filename
            return panel_bed

    def get_fastq_path(self, link_object: models.FamilySample) -> str:
        """Returns path to the concatenated FASTQ file of a sample"""
        file_collection = self.get_file_collection(sample_id=link_object.sample.internal_id)
        fastq_data = file_collection[0]
        linked_fastq_name = self.fastq_handler.FastqFileNameCreator.create(
            lane=fastq_data["lane"],
            flowcell=fastq_data["flowcell"],
            sample=link_object.sample.internal_id,
            read=fastq_data["read"],
            more={"undetermined": fastq_data["undetermined"]},
        )
        concatenated_fastq_name = self.fastq_handler.FastqFileNameCreator.get_concatenated_name(
            linked_fastq_name
        )
        concatenated_path = Path(
            self.balsamic_api.root_dir,
            link_object.family.internal_id,
            "fastq",
            concatenated_fastq_name,
        ).as_posix()
        return concatenated_path

    @staticmethod
    def get_sample_type(link_object: models.FamilySample) -> str:
        """Returns tissue type of a sample"""
        if link_object.sample.is_tumour:
            return "tumor"
        return "normal"

    @staticmethod
    def get_application_type(link_object: models.FamilySample) -> str:
        """Returns application type of a sample"""
        application_type = link_object.sample.application_version.application.prep_category
        return application_type

    def get_priority(self, case_id: str) -> str:
        """Returns priority for the case in clinical-db as text"""
        case_object = self.get_case_object(case_id)
        if case_object.high_priority:
            return "high"
        if case_object.low_priority:
            return "low"
        return "normal"

    def get_verified_bed(self, sample_data: dict, panel_bed: Path) -> Optional[str]:
        """ "Takes a dict with samples and attributes.
        Retrieves unique attributes for application type and target_bed.
        Verifies that those attributes are the same across multiple samples,
        where applicable.
        Verifies that the attributes are valid BALSAMIC attributes
        If application type requires bed, returns path to bed.

        Raises BalsamicStartError:
        - When application type invalid for balsamic
        - When multiple samples have different parameters
        - When bed file required for analysis, but is not set or cannot be retrieved.
        """
        application_types = set([v["application_type"].lower() for k, v in sample_data.items()])
        target_beds = set([v["target_bed"] for k, v in sample_data.items()])

        if not application_types.issubset(self.__BALSAMIC_APPLICATIONS):
            raise BalsamicStartError("Case application not compatible with BALSAMIC")
        if len(application_types) != 1:
            raise BalsamicStartError("Multiple application types found in LIMS")
        if not application_types.issubset(self.__BALSAMIC_BED_APPLICATIONS):
            if panel_bed:
                raise BalsamicStartError("Cannot set panel_bed for WGS sample!")
            return None
        if panel_bed:
            return panel_bed.as_posix()
        if len(target_beds) == 1:
            target_bed = target_beds.pop()
            if not target_bed:
                raise BalsamicStartError(
                    f"Application type {application_types.pop()} requires a bed file to be analyzed!"
                )
            return Path(self.balsamic_api.bed_path, target_bed).as_posix()
        raise BalsamicStartError(f"Too many BED versions in LIMS: {len(target_beds)}")

    @staticmethod
    def get_verified_tumor_path(sample_data: dict) -> str:
        """Takes a dict with samples and attributes, and returns the path
        of tumor sample.
        If the number of paths is exactly 1, the path is returned.
        Raises BalsamicStartError:
            When there are no tumor samples, or more than one tumor sample
        """
        tumor_paths = [
            val["concatenated_path"]
            for key, val in sample_data.items()
            if val["tissue_type"] == "tumor"
        ]
        if len(tumor_paths) != 1:
            raise BalsamicStartError(
                f"Invalid number of tumor samples: {len(tumor_paths)}, "
                f"BALSAMIC analysis requires exactly 1 tumor sample per case to run successfully!"
            )
        return tumor_paths[0]

    @staticmethod
    def get_verified_normal_path(sample_data: dict) -> Optional[str]:
        """Takes a dict with samples and attributes, and retrieves the path
        of normal sample.
        If the number of paths is exactly 1, the path is returned.
        If there are no paths, then the sample is not paired, and None is returned.
        Raises BalsamicStartError:
            When there is more than one normal sample.
        """
        normal_paths = [
            val["concatenated_path"]
            for key, val in sample_data.items()
            if val["tissue_type"] == "normal"
        ]
        if len(normal_paths) > 1:
            raise BalsamicStartError(
                f"Invalid number of normal samples: {len(normal_paths)}, only up to 1 allowed!!"
            )
        if len(normal_paths) == 0:
            return None
        return normal_paths[0]

    def get_verified_config_case_arguments(
        self,
        case_id: str,
        panel_bed: str,
    ) -> dict:
        """Takes a dictionary with per-sample parameters,
        validates them, and transforms into command line arguments
        Raises BalsamicStartError:
            When no samples associated with case are marked for BALSAMIC analysis
        """
        sample_data = self.get_sample_params(case_id=case_id, panel_bed=panel_bed)
        if len(sample_data) == 0:
            raise BalsamicStartError(f"{case_id} has no samples tagged for BALSAMIC analysis!")
        if panel_bed:
            if Path(f"{panel_bed}").is_file():
                panel_bed = Path(f"{panel_bed}")
            else:
                derived_panel_bed = Path(
                    self.balsamic_api.bed_path,
                    self.store.bed_version(panel_bed).filename,
                )
                if not derived_panel_bed.is_file():
                    raise BalsamicStartError(
                        f"{panel_bed} or {derived_panel_bed} are not valid paths to a BED file. "
                        f"Please provide absolute path to desired BED file or a valid bed shortname!"
                    )
                panel_bed = derived_panel_bed

        arguments = {
            "case_id": case_id,
            "normal": self.get_verified_normal_path(sample_data=sample_data),
            "tumor": self.get_verified_tumor_path(sample_data=sample_data),
            "panel_bed": self.get_verified_bed(sample_data=sample_data, panel_bed=panel_bed),
        }
        return arguments

    @staticmethod
    def print_sample_params(case_id: str, sample_data: dict) -> None:
        """Outputs a table of samples to be displayed in log"""

        LOG.info(f"Case {case_id} has following BALSAMIC samples:")
        LOG.info(
            "{:<20} {:<20} {:<20} {:<20}".format(
                "SAMPLE ID", "TISSUE TYPE", "APPLICATION", "BED VERSION"
            )
        )
        for key in sample_data:
            LOG.info(
                "{:<20} {:<20} {:<20} {:<20}".format(
                    key,
                    str(sample_data[key]["tissue_type"]),
                    str(sample_data[key]["application_type"]),
                    str(sample_data[key]["target_bed"]),
                )
            )
        LOG.info("")

    def get_sample_params(self, case_id: str, panel_bed: Optional[str]) -> dict:

        """Returns a dictionary of attributes for each sample in given family,
        where SAMPLE ID is used as key"""

        sample_data = {}
        for link_object in self.get_balsamic_sample_objects(case_id=case_id):
            sample_data[link_object.sample.internal_id] = {
                "tissue_type": self.get_sample_type(link_object),
                "concatenated_path": self.get_fastq_path(link_object),
                "application_type": self.get_application_type(link_object),
                "target_bed": self.resolve_target_bed(panel_bed=panel_bed, link_object=link_object),
            }
        self.print_sample_params(case_id=case_id, sample_data=sample_data)
        return sample_data

    def get_case_application_type(self, case_id: str) -> str:
        application_types = set(
            [
                self.get_application_type(link_object)
                for link_object in self.get_balsamic_sample_objects(case_id=case_id)
            ]
        )
        if application_types:
            return application_types.pop().lower()

    def resolve_target_bed(
        self, panel_bed: Optional[str], link_object: models.FamilySample
    ) -> Optional[str]:
        if panel_bed:
            return panel_bed
        if self.get_application_type(link_object) not in self.__BALSAMIC_BED_APPLICATIONS:
            return None
        return self.get_target_bed_from_lims(link_object)

    def parse_deliverables_report(self, case_id: str) -> list:
        """Parse BALSAMIC deliverables report, and return a list of files and their respective tags in bundle"""
        self.get_case_object(case_id=case_id)
        deliverables_file_path = self.get_deliverables_file_path(case_id=case_id, check_exists=True)
        sample_names = [
            x.sample.internal_id for x in self.get_balsamic_sample_objects(case_id=case_id)
        ]
        report_entries = dict(json.load(open(deliverables_file_path, "r")))["files"]
        bundle_files = []
        for entry in report_entries:
            tags = entry["tag"].split(",")
            for ind, tag in enumerate(tags):
                for sample_name in sample_names:
                    if sample_name in tag:
                        tags[ind] = sample_name

            bundle_file = {"path": entry["path"], "tags": tags, "archive": False}
            bundle_files.append(bundle_file)
        return bundle_files

    def upload_bundle_housekeeper(self, case_id: str) -> None:
        """ Add analysis bundle to Housekeeper """
        self.get_case_object(case_id=case_id)
        sample_config = self.get_config_path(case_id=case_id, check_exists=True)
        config_data = dict(json.load(open(sample_config, "r")))
        bundle_data = {
            "name": case_id,
            "created": dt.datetime.strptime(
                config_data["analysis"]["config_creation_date"],
                "%Y-%m-%d %H:%M",
            ),
            "version": config_data["analysis"]["BALSAMIC_version"],
            "files": self.parse_deliverables_report(case_id=case_id),
        }
        bundle_result = self.housekeeper_api.add_bundle(bundle_data=bundle_data)
        if not bundle_result:
            raise BundleAlreadyAddedError("Bundle already added to Housekeeper!")
        bundle_object, bundle_version = bundle_result
        self.housekeeper_api.include(bundle_version)
        self.housekeeper_api.add_commit(bundle_object, bundle_version)
        LOG.info(
            f"Analysis successfully stored in Housekeeper: {case_id} : {bundle_version.created_at}"
        )

    def upload_analysis_statusdb(self, case_id: str) -> None:
        """ Add Analysis bundle to StatusDB """
        case_object = self.get_case_object(case_id=case_id)
        sample_config = self.get_config_path(case_id=case_id, check_exists=True)
        config_data = dict(json.load(open(sample_config, "r")))
        analysis_start = dt.datetime.strptime(
            config_data["analysis"]["config_creation_date"], "%Y-%m-%d %H:%M"
        )
        case_object.action = None
        new_analysis = self.store.add_analysis(
            pipeline=Pipeline.BALSAMIC,
            version=config_data["analysis"]["BALSAMIC_version"],
            started_at=analysis_start,
            completed_at=dt.datetime.now(),
            primary=(len(case_object.analyses) == 0),
        )
        new_analysis.family = case_object
        self.store.add_commit(new_analysis)
        LOG.info(f"Analysis successfully stored in ClinicalDB: {case_id} : {analysis_start}")

    def family_has_correct_number_tumor_normal_samples(self, case_id: str) -> bool:
        """Evaluates if a case has exactly one tumor and up to one normal sample in ClinicalDB.
        This check is only applied to filter jobs which start automatically"""

        query = (
            self.store.Sample.query.join(models.Family.links, models.FamilySample.sample)
            .filter(models.Family.internal_id == case_id)
            .filter(models.Family.data_analysis == str(Pipeline.BALSAMIC))
        )

        return all(
            [
                len(query.filter(models.Sample.is_tumour == False).all()) <= 1,
                len(query.filter(models.Sample.is_tumour == True).all()) == 1,
            ]
        )

    def get_analyses_to_clean(self, before_date: dt.datetime = dt.datetime.now()) -> list:
        """Retrieve a list of analyses for cleaning created before certain date"""
        analyses_to_clean = self.store.analyses_to_clean(
            pipeline=Pipeline.BALSAMIC, before=before_date
        )
        return analyses_to_clean.all()

    def get_cases_to_analyze(self) -> list:
        """Retrieve a list of balsamic cases without analysis,
        where samples have enough reads to be analyzed"""
        cases_to_analyze = []
        for case_object in self.store.cases_to_analyze(pipeline=Pipeline.BALSAMIC, threshold=0.75):
            if self.family_has_correct_number_tumor_normal_samples(case_object.internal_id):
                cases_to_analyze.append(case_object.internal_id)
        return cases_to_analyze

    def get_cases_to_store(self) -> list:
        """Retrieve a list of cases where analysis finished successfully,
        and is ready to be stored in Housekeeper"""
        cases_to_store = []
        for case_object in self.store.cases_to_store(pipeline=Pipeline.BALSAMIC):
            case_id = case_object.internal_id
            if Path(self.get_analysis_finish_path(case_id=case_id)).exists():
                cases_to_store.append(case_id)
        return cases_to_store



"""Module for mocking out the HK api in CG"""

import datetime
import logging
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import List

ROOT_PATH = tempfile.TemporaryDirectory().name

LOG = logging.getLogger(__name__)


def calculate_checksum(path):
    """Calculate the checksum for a file"""
    _ = path
    return "asdjkfghasdkfj"


class MockTag:
    """Mocks a hk tag object"""

    def __init__(self, **kwargs):
        """Init a tag mock. Defaults name to 'vcf'"""
        self.id = kwargs.get("id", 1)
        self.name = kwargs.get("name", "vcf")
        self.category = kwargs.get("category")
        self.created_at = kwargs.get("created_at", datetime.datetime.now())

    def __repr__(self):
        return (
            f"MockTag:id={self.id}, name={self.name}, category={self.category},"
            f"created_at={self.created_at}"
        )


class MockFile:
    """Mocks a hk file object"""

    def __init__(self, **kwargs):
        """Init a file mock"""
        self.id = kwargs.get("id", 1)
        self.path = kwargs.get("path", "a_file")
        self.checksum = calculate_checksum(self.path)
        self.to_archive = kwargs.get("to_archive", False)

        self.version_id = kwargs.get("version_id", 1)
        self.tags = kwargs.get("tags", [MockTag()])

        self.app_root = Path(kwargs.get("root_path", ROOT_PATH))

    @property
    def full_path(self):
        """Return the full path to the file."""
        return str(self.path)

    @property
    def is_included(self):
        """Check if the file is included in Housekeeper."""
        return str(self.app_root) in self.full_path

    def delete(self):
        """Mock delete functions"""
        return True

    def __repr__(self):
        return f"MockFile:id={self.id}, path={self.path}, to_archive={self.to_archive}"


class MockVersion:
    """Mocks a hk version object"""

    def __init__(self, **kwargs):
        """Init a version mock"""
        self.id = kwargs.get("id", 1)
        self.created_at = kwargs.get("created_at", datetime.datetime.now())
        self.expires_at = kwargs.get("expires_at")
        self.included_at = kwargs.get("included_at")
        self.removed_at = kwargs.get("removed_at")

        self.archived_at = kwargs.get("archived_at")
        self.archive_path = kwargs.get("archive_path")
        self.archive_checksum = kwargs.get("archive_checksum")

        self.bundle = kwargs.get("bundle")
        self.bundle_name = kwargs.get("bundle", "bundle_name")
        self.bundle_id = kwargs.get("bundle_id", 1)

        self.files = kwargs.get("files", [])

        self.app_root = kwargs.get("root_path", ROOT_PATH)

    @property
    def relative_root_dir(self):
        """Build the relative root dir path for the bundle version."""
        return Path(self.bundle_name) / str(self.created_at.date())

    @property
    def full_path(self):
        """Returns the full path of the bundle"""
        return Path(self.app_root) / self.bundle.name / str(self.created_at.date())

    def __repr__(self):
        return f"MockVersion:id={self.id}, created_at={self.created_at}, files={self.files}"


class EnhancedList(list):
    """Create a list that mocks the behaviour of a query result"""

    def __init__(self):
        super(EnhancedList, self).__init__()

    def first(self):
        """Mock the first method """
        if len(self) == 0:
            return None
        return self[0]

    def count(self):
        """Mock the count method"""
        return len(self)

    def all(self):
        """Mock the all method"""
        return self


class MockBundle:
    """Mocks a hk bundle object"""

    def __init__(self, **kwargs):
        """Init a bundle mock"""
        self.id = kwargs.get("id", 1)
        self.name = kwargs.get("name", "yellowhog")
        self.created_at = kwargs.get("created_at", datetime.datetime.now())
        self.versions = kwargs.get("versions", [MockVersion(bundle=self)])

    def __repr__(self):
        return f"MockBundle:id={self.id}, name={self.name}, versions={self.versions}"


class MockHousekeeperAPI:
    """Mocks all the behaviour of the housekeeper API"""

    def __init__(self, config=None):
        self._version_obj = None
        self._bundle_obj = None
        self._files = EnhancedList()
        self._tags = EnhancedList()
        self._bundles = EnhancedList()
        self._id_counter = 1
        self._file_added = False
        self._file_included = False
        self._tags_matter = False
        self._last_version = True
        # Add tags here if there should be missing files
        self._missing_tags = set()
        if not config:
            config = {
                "housekeeper": {
                    "database": "sqlite:///:memory:",
                    "root": str(ROOT_PATH),
                }
            }
        self._database = config.get("housekeeper", {}).get("database")
        self.root_path = config.get("housekeeper", {}).get("root", str(ROOT_PATH))

    # Mock specific functions
    def add_missing_tag(self, tag_name: str):
        """Add a missing tag"""
        self._missing_tags.add(tag_name)

    def set_missing_last_version(self):
        """Make sure that no version is returned"""
        self._last_version = False

    def is_file_included(self) -> bool:
        """Return true if any file has been included"""
        return self._file_included

    def is_file_added(self) -> bool:
        """Return true if any file has been added"""
        return self._file_added

    def tag_exists(self, tag_name) -> bool:
        """Return true if a tag has been added"""
        for tag_obj in self._tags:
            if tag_obj.name == tag_name:
                return True
        return False

    def file_exists(self, file_path) -> bool:
        """Return true if a file has been added"""
        file_name = Path(file_path).name
        for file_obj in self._files:
            if Path(file_obj.path).name == file_name:
                return True
        return False

    def update_id_counter(self):
        """Increment id counter"""
        self._id_counter += 1

    # Mocked functions from original API
    def add_bundle(self, bundle_data):
        """ Build a new bundle version of files """
        bundle_obj = self.new_bundle(name=bundle_data["name"], created_at=bundle_data["created"])

        version_obj = self.new_version(
            created_at=bundle_data["created"], expires_at=bundle_data.get("expires")
        )

        tag_names = set(
            tag_name for file_data in bundle_data["files"] for tag_name in file_data["tags"]
        )
        tag_map = self._build_tags(tag_names)
        for file_data in bundle_data["files"]:
            if isinstance(file_data["path"], str):
                paths = [file_data["path"]]
            else:
                paths = file_data["path"]
            for path in paths:
                tags = [tag_map[tag_name] for tag_name in file_data["tags"]]

                new_file = self.new_file(path, to_archive=file_data["archive"], tags=tags)
                self._files.append(new_file)
                self._file_added = True
                version_obj.files.append(new_file)
        version_obj.bundle_obj = bundle_obj
        bundle_obj.versions = version_obj

        return bundle_obj, version_obj

    def _build_tags(self, tag_names: List[str]) -> dict:
        """Build a list of tag objects."""
        tags = {}
        for tag_name in tag_names:
            if self.tag_exists(tag_name):
                tag_obj = self.tag(tag_name)
            else:
                tag_obj = self.new_tag(tag_name)
                self._tags.append(tag_obj)
            tags[tag_name] = tag_obj
        return tags

    def tag(self, name: str):
        """ Fetch a tag """
        for tag_obj in self._tags:
            if tag_obj.name == name:
                return tag_obj
        return None

    def bundle(self, name: str) -> MockBundle:
        """ Fetch a bundle """
        if name:
            for bundle_obj in self.bundles():
                if bundle_obj.name == name:
                    return bundle_obj
        return self._bundle_obj

    def bundles(self):
        """ Fetch bundles """
        return self._bundles

    def new_bundle(self, name: str, created_at: datetime.datetime = None):
        """ Create a new file bundle """
        self.update_id_counter()
        bundle_obj = MockBundle(id=self._id_counter, name=name, created_at=created_at)
        self._bundle_obj = bundle_obj
        self._bundles.append(bundle_obj)
        return bundle_obj

    def version(self, *args, **kwargs):
        """ Fetch a version """
        return self._version_obj

    def files(self, *args, **kwargs):
        """
        Fetch files.
        If it has been specified that some files should be missing return empty list
        """
        tags = set(kwargs.get("tags", []))
        if tags.intersection(self._missing_tags):
            return EnhancedList()
        return self._files

    def new_tag(self, name: str, category: str = None):
        """ Create a new tag """
        self.update_id_counter()
        tag_obj = MockTag(id=self._id_counter, name=name, category=category)
        if not self.tag_exists(name):
            self._tags.append(tag_obj)

        return tag_obj

    def add_tag(self, name: str, category: str = None):
        """ Add a tag to the database """
        tag_obj = self.new_tag(name, category)
        if not self.tag_exists(name):
            self._tags.append(tag_obj)
        return tag_obj

    def new_version(self, created_at: datetime.datetime, expires_at: datetime.datetime = None):
        """ Create a new bundle version """
        self.update_id_counter()
        created_at = created_at or datetime.datetime.now()
        expires_at = expires_at or datetime.datetime.now()
        version_obj = MockVersion(id=self._id_counter, created_at=created_at, expires_at=expires_at)
        self._version_obj = version_obj
        return version_obj

    def add_version(
        self,
        version_obj: MockVersion = None,
        created_at: datetime.datetime = None,
        expires_at: datetime.datetime = None,
    ):
        """ Create a new bundle version """
        if not version_obj:
            version_obj = self.new_version(created_at, expires_at)
        self._version_obj = version_obj
        return version_obj

    def new_file(
        self,
        path: str,
        checksum: str = None,
        to_archive: bool = False,
        tags: list = [],
    ):
        """ Create a new file """
        self.update_id_counter()
        mocked_file = MockFile(
            id=self._id_counter,
            path=path,
            checksum=checksum,
            to_archive=to_archive,
            tags=tags,
        )
        if not self.file_exists(path):
            self._files.append(mocked_file)
        self._file_added = True
        return mocked_file

    def add_commit(self, *args, **kwargs):
        """ Wrap method in Housekeeper Store """
        return True

    def commit(self):
        """ Wrap method in Housekeeper Store """
        return True

    def include(self, *args, **kwargs):
        """Call the include version function to import related assets."""
        self._file_included = True

    def include_file(self, *args, **kwargs):
        """Call the include version function to import related assets."""
        self._file_included = True

    def last_version(self, *args, **kwargs):
        """Gets the latest version of a bundle"""
        if self._last_version is False:
            return None
        return self._version_obj

    def get_root_dir(self):
        """Returns the root dir of Housekeeper"""
        return self.root_path

    def get_files(self, *args, **kwargs):
        """Fetch all the files in housekeeper, optionally filtered by bundle and/or tags and/or
        version

        Returns:
            iterable(hk.Models.File)
        """
        return self.files(*args, **kwargs)

    def add_file(self, path, version_obj, tags, to_archive=False):
        """Add a file to housekeeper."""
        tags = tags or []
        if isinstance(tags, str):
            tags = [tags]
        for tag_name in tags:
            if not self.tag(tag_name):
                self.add_tag(tag_name)

        new_file = self.new_file(
            path=str(Path(path).absolute()),
            to_archive=to_archive,
            tags=[self.tag(tag_name) for tag_name in tags],
        )
        if not version_obj:
            version_obj = self.new_version(created_at=datetime.datetime.now())
        new_file.version = version_obj
        if not self.file_exists(path):
            self._files.append(new_file)
        self._file_added = True
        return new_file

    @staticmethod
    def checksum(path):
        """Calculate the checksum"""
        return calculate_checksum(path)

    def initialise_db(self):
        """Create all tables in the store."""

    def destroy_db(self):
        """Drop all tables in the store"""

    @contextmanager
    def session_no_autoflush(self):
        """ Wrap property in Housekeeper Store """
        yield True

    def __repr__(self):
        return f"HousekeeperMockAPI:version_obj={self._version_obj}"


if __name__ == "__main__":
    hk_api = MockHousekeeperAPI(config={})



"""Tests for the compress fastq cli"""

import logging

from cg.cli.compress.fastq import fastq_cmd


def test_compress_fastq_cli_no_family(compress_context, cli_runner, caplog):
    """Test to run the compress command with a database without samples"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, [], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert "Individuals in 0 (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_case_id_no_family(compress_context, cli_runner, case_id, caplog):
    """Test to run the compress command when no families are found"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--case-id", case_id], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Could not find case {case_id}" in caplog.text


def test_real_compress_fastq_cli_one_family(
    real_populated_compress_context, cli_runner, case_id, caplog
):
    """Test to run the compress command there is an existing family"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a family
    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--case-id", case_id], obj=real_populated_compress_context)

    # THEN assert the program exits with a non zero exit status
    assert res.exit_code == 0
    # THEN assert it was communicated that one family was compressed
    assert f"Individuals in 1 (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_case_id(populated_compress_context, cli_runner, case_id, caplog):
    """Test to run the compress command with a specified case id"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a family

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--case-id", case_id], obj=populated_compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Individuals in 1 (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_multiple_family(
    populated_multiple_compress_context, cli_runner, caplog
):
    """Test to run the compress command with multiple families"""
    compress_context = populated_multiple_compress_context
    caplog.set_level(logging.DEBUG)
    # GIVEN a database with multople families
    nr_cases = sum(1 for i in compress_context["status_db"].families())
    assert nr_cases > 1

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--number-of-conversions", nr_cases], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Individuals in {nr_cases} (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_multiple_set_limit(
    populated_multiple_compress_context, cli_runner, caplog
):
    """Test to run the compress command with multiple families and use a limit"""
    compress_context = populated_multiple_compress_context
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with more families than the limit
    nr_cases = sum(1 for i in compress_context["status_db"].families())
    limit = 5
    assert nr_cases > limit

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--number-of-conversions", limit], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated no more than the limited number of cases was compressed
    assert f"Individuals in {limit} (completed) cases where compressed" in caplog.text



"""Fixtures for testing balsamic app"""
import string

import pytest


@pytest.fixture
def valid_fastq_filename_pattern():
    """the pattern balsamic file names should match"""
    # 'xxx_R_1.fastq.gz and xxx_R_2.fastq.gz'
    return r"^.+_R_[1-2]{1}\.fastq.gz$"


def _full_content():
    """The content the files are made of"""
    return string.ascii_letters


@pytest.fixture
def files_content(tmpdir):
    """The content the files are made of"""
    return _full_content()[0 : len(_simple_files(tmpdir))]


@pytest.fixture
def content_r1(tmpdir):
    """The content of concatenated r1 """
    # return full_content[0:len(simple_files) // 2]
    return "".join(simple(tmpdir)["content_r1"])


@pytest.fixture
def content_r2(tmpdir):
    """The content of concatenated r2 """
    # return full_content[len(simple_files) // 2:len(simple_files)]
    return "".join(simple(tmpdir)["content_r2"])


def simple(tmpdir):
    """Creates a dict with the data to use in the tests"""
    flowcells = [1, 2, 3, 4, 5, 6, 7, 8]
    lanes = [1, 2, 3]
    reads = [1, 2]

    _simple = {
        "files": [],
        "content_r1": [],
        "content_r2": [],
        "data": [],
        "data_reversed": [],
    }
    i = 0

    for read in reads:
        for flowcell in flowcells:
            for lane in lanes:
                content = _full_content()[i]
                file_path = create_file(tmpdir, flowcell, lane, read, content)

                _simple["files"].append(file_path)

                if read == 1:
                    _simple["content_r1"].append(content)
                else:
                    _simple["content_r2"].append(content)

                data = create_file_data(file_path, flowcell, lane, read)
                _simple["data"].append(data)
                _simple["data_reversed"].insert(0, data)
                i += 1

    return _simple


def _simple_files(tmpdir):
    """"Some files to test with"""
    return simple(tmpdir)["files"]


@pytest.fixture
def simple_files(tmpdir):
    """"Some files to test with"""
    return _simple_files(tmpdir)


@pytest.fixture
def simple_files_data(tmpdir):
    """Data for link method"""
    return simple(tmpdir)["data"]


@pytest.fixture
def simple_files_data_reversed(tmpdir):
    """Data for link method"""
    return simple(tmpdir)["data_reversed"]


def create_file(tmpdir, flowcell, lane, read, file_content):
    """actual file on disk"""

    file_name = f"S1_FC000{flowcell}_L00{lane}_R_{read}.fastq.gz"
    file_path = tmpdir / file_name
    file_path.write(file_content)
    return file_path


def create_file_data(file_path, flowcell, lane, read):
    """meta data about a file on disk"""
    data = {
        "path": file_path,
        "lane": lane,
        "flowcell": flowcell,
        "read": read,
        "undetermined": False,
    }
    return data


@pytest.fixture
def cg_config(tmpdir):
    """mock relevant parts of a cg-config"""
    return {"balsamic": {"root": tmpdir}}


@pytest.fixture
def link_family():
    """mock family name"""
    return "family"


@pytest.fixture
def link_sample():
    """mock sample name"""
    return "sample"



"""cg module for cleaning databases and files"""
import logging
import shutil
from datetime import datetime
from pathlib import Path

import click
from dateutil.parser import parse as parse_date

from cg.apps.balsamic.api import BalsamicAPI
from cg.apps.balsamic.fastq import FastqHandler
from cg.apps.crunchy import CrunchyAPI
from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.store import Store


LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def clean(context):
    """Clean up processes"""
    context.obj["status_db"] = Store(context.obj["database"])
    context.obj["housekeeper_api"] = HousekeeperAPI(context.obj)
    context.obj["trailblazer_api"] = TrailblazerAPI(context.obj)
    context.obj["scout_api"] = ScoutAPI(context.obj)
    context.obj["crunchy_api"] = CrunchyAPI(context.obj)
    context.obj["lims_api"] = LimsAPI(context.obj)

    context.obj["BalsamicAnalysisAPI"] = BalsamicAnalysisAPI(
        balsamic_api=BalsamicAPI(context.obj),
        store=context.obj["status_db"],
        housekeeper_api=context.obj["housekeeper_api"],
        fastq_handler=FastqHandler(context.obj),
        lims_api=context.obj["lims_api"],
        trailblazer_api=context.obj["trailblazer_api"],
    )
    context.obj["MipAnalysisAPI"] = MipAnalysisAPI(
        db=context.obj["status_db"],
        hk_api=context.obj["housekeeper_api"],
        tb_api=context.obj["trailblazer_api"],
        scout_api=context.obj["scout_api"],
        lims_api=context.obj["lims_api"],
        script=context.obj["mip-rd-dna"]["script"],
        pipeline=context.obj["mip-rd-dna"]["pipeline"],
        conda_env=context.obj["mip-rd-dna"]["conda_env"],
        root=context.obj["mip-rd-dna"]["root"],
    )


@clean.command("balsamic-run-dir")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("case_id")
@click.pass_context
def balsamic_run_dir(context, yes, case_id, dry_run: bool = False):
    """Remove Balsamic run directory"""

    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    case_object = balsamic_analysis_api.get_case_object(case_id)
    if not case_object:
        LOG.warning(f"{case_id} not found!")
        raise click.Abort()

    analysis_obj = case_object.analyses[0] if case_object.analyses else None
    if not analysis_obj:
        LOG.error("%s: analysis not found", case_id)
        raise click.Abort()

    analysis_path = Path(balsamic_analysis_api.get_case_path(case_id))

    if dry_run:
        LOG.info("Would have deleted: %s", analysis_path)
        return EXIT_SUCCESS

    if yes or click.confirm(f"Are you sure you want to remove all files in {analysis_path}?"):
        if not analysis_path.exists():
            LOG.warning("could not find: %s", analysis_path)
            return EXIT_FAIL
        if analysis_path.is_symlink():
            LOG.warning(
                "Will not automatically delete symlink: %s, delete it manually", analysis_path
            )
            return EXIT_FAIL

        try:
            shutil.rmtree(analysis_path, ignore_errors=True)
            LOG.info(f"Cleaned {analysis_path}")
        except Exception as e:
            LOG.warning(
                f" Directory {analysis_path} will not be deleted due to unexpected error - {e}!"
            )

        analysis_obj.cleaned_at = datetime.now()
        balsamic_analysis_api.store.commit()


@clean.command("mip-run-dir")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("case_id")
@click.pass_context
def mip_run_dir(context, yes, case_id, dry_run: bool = False):
    """Remove MIP run directory"""

    mip_analysis_api = context.obj["MipAnalysisAPI"]
    case_obj = mip_analysis_api.get_case_object(case_id)

    if case_obj is None:
        LOG.error("%s: family not found", case_id)
        raise click.Abort()

    analysis_path = mip_analysis_api.get_case_output_path(case_id=case_id)
    if not analysis_path.exists():
        LOG.error(f"No analysis directory found for {case_id}")
        raise click.Abort()

    if dry_run:
        LOG.info(f"Cleaning case {case_id} : Would have deleted contents of {analysis_path}")
        return

    try:
        if mip_analysis_api.is_latest_analysis_ongoing(case_id):
            LOG.warning(f"Analysis for case {case_id} is still ongoing!")

        if yes or click.confirm(f"Are you sure you want to remove {case_id}?"):
            shutil.rmtree(analysis_path, ignore_errors=True)
            LOG.info(f"Cleaning case {case_id} : Deleted contents of {analysis_path}")
            mip_analysis_api.mark_analyses_deleted(case_id)
            for analysis_obj in case_obj.analyses:
                analysis_obj.cleaned_at = analysis_obj.cleaned_at or datetime.now()
            mip_analysis_api.db.commit()
    except Exception as error:
        LOG.error(f"{case_id}: {error}")
        raise click.Abort()


@clean.command("hk-alignment-files")
@click.argument("bundle")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Show files that would be cleaned")
@click.pass_context
def hk_alignment_files(context, bundle, yes: bool = False, dry_run: bool = False):
    """Clean up alignment files in Housekeeper bundle"""
    files = []
    for tag in ["bam", "bai", "bam-index", "cram", "crai", "cram-index"]:
        files.extend(context.obj["housekeeper_api"].get_files(bundle=bundle, tags=[tag]))
    for file_obj in files:
        if file_obj.is_included:
            question = f"{bundle}: remove file from file system and database: {file_obj.full_path}"
        else:
            question = f"{bundle}: remove file from database: {file_obj.full_path}"

        if yes or click.confirm(question):
            file_name = file_obj.full_path
            if file_obj.is_included and Path(file_name).exists():
                if not dry_run:
                    Path(file_name).unlink()

            if not dry_run:
                file_obj.delete()
                context.obj["housekeeper_api"].commit()
                click.echo(f"{file_name} deleted")


@clean.command("scout-finished-cases")
@click.option(
    "--days-old",
    type=int,
    default=300,
    help="Clean alignment files with analysis dates older then given number of days",
)
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.pass_context
def scout_finished_cases(context, days_old: int, yes: bool = False, dry_run: bool = False):
    """Clean up of solved and archived scout cases"""
    bundles = []
    for status in "archived", "solved":
        cases = context.obj["scout_api"].get_cases(status=status, reruns=False)
        cases_added = 0
        for case in cases:
            x_days_ago = datetime.now() - case.get("analysis_date")
            if x_days_ago.days > days_old:
                bundles.append(case.get("_id"))
                cases_added += 1
        LOG.info(f"{cases_added} cases marked for bam removal")

    for bundle in bundles:
        context.invoke(hk_alignment_files, bundle=bundle, yes=yes, dry_run=dry_run)


@clean.command("hk-past-files")
@click.option("-c", "--case-id", type=str)
@click.option("-t", "--tags", multiple=True)
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.pass_context
def hk_past_files(context, case_id, tags, yes, dry_run):
    """ Remove files found in older housekeeper bundles """
    if case_id:
        cases = [context.obj["status_db"].family(case_id)]
    else:
        cases = context.obj["status_db"].families()
    for case in cases:
        case_id = case.internal_id
        last_version = context.obj["housekeeper_api"].last_version(bundle=case_id)
        if not last_version:
            continue
        last_version_file_paths = [
            Path(hk_file.full_path)
            for hk_file in context.obj["housekeeper_api"].get_files(
                bundle=case_id, tags=None, version=last_version.id
            )
        ]
        LOG.info("Searching %s bundle for outdated files", case_id)
        hk_files = []
        for tag in tags:
            hk_files.extend(context.obj["housekeeper_api"].get_files(bundle=case_id, tags=[tag]))
        for hk_file in hk_files:
            file_path = Path(hk_file.full_path)
            if file_path in last_version_file_paths:
                continue
            LOG.info("Will remove %s", file_path)
            if yes or click.confirm("Do you want to remove this file?"):
                if not dry_run:
                    hk_file.delete()
                    context.obj["housekeeper_api"].commit()
                    if file_path.exists():
                        file_path.unlink()
                    LOG.info("File removed")


@clean.command("balsamic-past-run-dirs")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("before_str")
@click.pass_context
def balsamic_past_run_dirs(context, before_str: str, yes: bool = False, dry_run: bool = False):
    """Clean up of "old" Balsamic case run dirs"""

    before = parse_date(before_str)
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    possible_cleanups = balsamic_analysis_api.get_analyses_to_clean(before_date=before)
    LOG.info(f"Cleaning all analyses created before {before}")

    # for all analyses
    for analysis in possible_cleanups:
        case_id = analysis.family.internal_id

        # call clean
        LOG.info(f"Cleaning Balsamic output for {case_id}")
        try:
            context.invoke(balsamic_run_dir, yes=yes, case_id=case_id, dry_run=dry_run)
        except click.Abort:
            continue

    LOG.info("Done cleaning Balsamic output")


@clean.command("mip-past-run-dirs")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.argument("before_str")
@click.pass_context
def mip_past_run_dirs(
    context: click.Context, before_str: str, yes: bool = False, dry_run: bool = False
):
    """Clean up of "old" MIP case run dirs"""
    mip_analysis_api = context.obj["MipAnalysisAPI"]
    before = parse_date(before_str)
    old_analyses = mip_analysis_api.get_analyses_to_clean(before=before)
    for status_analysis in old_analyses:
        case_id = status_analysis.family.internal_id
        try:
            context.invoke(
                mip_run_dir,
                yes=yes,
                case_id=case_id,
                dry_run=dry_run,
            )
        except click.Abort:
            continue



"""Import all constants here for easy access"""

from .tags import HK_TAGS, HK_FASTQ_TAGS, MIP_DNA_TAGS, MIP_RNA_TAGS, MICROSALT_TAGS
from .symbols import SPACE, SINGLE_QUOTE, NO_PARENT
from .process import EXIT_FAIL, EXIT_SUCCESS, RETURN_SUCCESS
from .compression import (
    FASTQ_FIRST_READ_SUFFIX,
    FASTQ_SECOND_READ_SUFFIX,
    FASTQ_DELTA,
    SPRING_SUFFIX,
)
from .sample_sources import METAGENOME_SOURCES, ANALYSIS_SOURCES
from .gene_panel import MASTER_LIST
from .paths import TMP_DIR
from .priority import PRIORITY_MAP, PRIORITY_OPTIONS, REV_PRIORITY_MAP

from .constants import (
    CAPTUREKIT_OPTIONS,
    CAPTUREKIT_CANCER_OPTIONS,
    COLLABORATORS,
    CONTAINER_OPTIONS,
    DEFAULT_CAPTURE_KIT,
    DataDelivery,
    COMBOS,
    FAMILY_ACTIONS,
    FLOWCELL_STATUS,
    Pipeline,
    PREP_CATEGORIES,
    SEX_OPTIONS,
    STATUS_OPTIONS,
)



"""Contains class to calculate values on samples"""
from cg.store import models


class SampleCalculator:
    """Calculates on samples"""

    @staticmethod
    def calculate_processing_days(sample: models.Sample):
        """Calculates the days from received to delivered"""
        if sample.received_at and sample.delivered_at:
            return (sample.delivered_at - sample.received_at).days

        return None



"""Fixtures for crunchy api"""
import copy
import json
import logging
from pathlib import Path
from typing import List

import pytest

from cg.apps.crunchy.models import CrunchyFileSchema
from cg.models import CompressionData

LOG = logging.getLogger(__name__)


class MockProcess:
    """Mock a process"""

    def __init__(self, process_name: str):
        """Inititalise mock"""
        self.stderr = ""
        self.stdout = ""
        self.base_call = [process_name]
        self.process_name = process_name

    def run_command(self, parameters: List):
        """Mock the run process method"""
        command = copy.deepcopy(self.base_call)
        if parameters:
            command.extend(parameters)

        LOG.info("Running command %s", " ".join(command))


# File fixtures


@pytest.fixture(scope="function", name="file_schema")
def fixture_file_schema():
    """Return a instance of the file schema that describes content of SPRING metadata"""
    return CrunchyFileSchema()


@pytest.fixture(name="real_spring_metadata_path")
def fixture_real_spring_metadata_path(fixtures_dir):
    """Return the path to a SPRING metadata file"""
    return fixtures_dir / "apps" / "crunchy" / "spring_metadata.json"


@pytest.fixture(name="spring_metadata")
def fixture_spring_metadata(compression_object: CompressionData) -> List[dict]:
    """Return meta data information"""
    first_read = compression_object.fastq_first
    second_read = compression_object.fastq_second
    spring_path = compression_object.spring_path
    metadata = [
        {
            "path": str(first_read),
            "file": "first_read",
            "checksum": "checksum_first_read",
            "algorithm": "sha256",
        },
        {
            "path": str(second_read),
            "file": "second_read",
            "checksum": "checksum_second_read",
            "algorithm": "sha256",
        },
        {"path": str(spring_path), "file": "spring"},
    ]
    return metadata


@pytest.fixture(scope="function", name="spring_metadata_file")
def fixture_spring_metadata_file(
    compression_object: CompressionData, spring_metadata: List[dict]
) -> Path:
    """Return the path to a populated SPRING metadata file"""
    metadata_path = compression_object.spring_metadata_path

    with open(metadata_path, "w") as outfile:
        outfile.write(json.dumps(spring_metadata))

    return metadata_path


@pytest.fixture(scope="function", name="sbatch_process")
def fixture_sbatch_process():
    """Return a mocked process object"""
    return MockProcess("sbatch")


@pytest.fixture(scope="function", name="fastq_first_file")
def fixture_fastq_first_file(fastq_first_path):
    """Creates an existing FASTQ path"""
    fastq_first_path.touch()
    return fastq_first_path


@pytest.fixture(scope="function", name="fastq_second_file")
def fixture_fastq_second_file(fastq_second_path):
    """Creates an existing FASTQ path"""
    fastq_second_path.touch()
    return fastq_second_path


@pytest.fixture(scope="function", name="spring_file")
def fixture_spring_file(spring_path):
    """Creates an existing SPRING file"""
    spring_path.touch()
    return spring_path


@pytest.fixture(scope="function", name="fastq_paths")
def fixture_fastq_paths(fastq_first_path, fastq_second_path):
    """Creates FASTQ paths"""
    return {
        "fastq_first_path": fastq_first_path,
        "fastq_second_path": fastq_second_path,
    }


@pytest.fixture(scope="function", name="existing_fastq_paths")
def fixture_existing_fastq_paths(fastq_first_file, fastq_second_file):
    """Creates existing FASTQ paths"""
    return {
        "fastq_first_path": fastq_first_file,
        "fastq_second_path": fastq_second_file,
    }


@pytest.fixture(scope="function", name="compressed_fastqs")
def fixture_compressed_fastqs(existing_fastq_paths, spring_file, spring_metadata_file):
    """Creates FASTQs with corresponding SPRING and FLAG"""
    fastq_paths = existing_fastq_paths
    flag_path = spring_metadata_file

    assert spring_file.exists()
    assert flag_path.exists()
    return fastq_paths



from typing import Any, Union

import pytest
import copy
import datetime as dt

from cg.constants import Pipeline
from cg.meta.workflow.mip import MipAnalysisAPI
from pathlib import Path
from ruamel.yaml import YAML

from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.store import Store
from cg.apps.hk import HousekeeperAPI
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.tb_mock import MockTB

from trailblazer.store.models import Analysis as tb_Analysis


@pytest.fixture
def mip_lims():
    mip_lims = MockLimsAPI({})
    return mip_lims


@pytest.fixture
def mock_root_folder(project_dir):
    return Path(project_dir, "cases").as_posix()


@pytest.fixture(name="mip_case_ids")
def fixture_mip_case_ids() -> dict:
    """Dictionary of case ids, connected samples, their name and if they should fail (textbook or not)"""

    case_ids = {
        "yellowhog": {
            "textbook": True,
            "name": "F0000001",
            "internal_id": "ACC00000",
        },
        "purplesnail": {
            "textbook": False,
            "name": "F0000003",
            "internal_id": "ACC00000",
        },
        "bluezebra": {
            "textbook": True,
            "name": "F0000002",
            "internal_id": "ACC00000",
        },
    }

    return case_ids


@pytest.fixture(name="mip_case_dirs")
def fixture_mip_case_dirs(mip_case_ids: dict, project_dir: Path) -> dict:
    """Create case directories and return a dictionary of the tmpdir paths"""

    mip_case_dirs = {}

    for case in mip_case_ids:
        case_dir = Path(project_dir / "cases" / case)
        case_dir.mkdir(exist_ok=True, parents=True)
        mip_case_dirs[case] = case_dir

    return mip_case_dirs


@pytest.fixture(name="mip_deliverables")
def fixture_mip_deliverables(
    mip_case_ids: dict,
    mip_case_dirs: dict,
    mip_deliverables_file: Path,
) -> dict:
    """Create deliverables for mip store testing"""

    yaml = YAML()

    # Create a dict of paths to the deliverables to be used in later stages
    deliverables_paths = {}

    # Load general deliverables file
    with open(mip_deliverables_file, "r") as mip_dna_deliverables:
        mip_deliverables = yaml.load(mip_dna_deliverables)

    for case in mip_case_ids:
        if mip_case_ids[case]["textbook"]:
            case_specific_deliverables = copy.deepcopy(mip_deliverables)

            for file in case_specific_deliverables["files"]:
                if file["path_index"]:
                    file["path_index"] = str(mip_case_dirs[case] / Path(file["path_index"]).name)
                    # Touch the tmp file
                    open(file["path_index"], "a").close()
                file["path"] = str(mip_case_dirs[case] / Path(file["path"]).name)
                # Touch the tmp file
                open(file["path"], "a").close()

            case_deliverables_path = mip_case_dirs[case] / f"{case}_deliverables.yaml"
            with open(case_deliverables_path, "w") as fh:
                yaml.dump(case_specific_deliverables, fh)
            deliverables_paths[case] = case_deliverables_path

        elif not mip_case_ids[case]["textbook"]:
            case_deliverables_path = mip_case_dirs[case] / f"{case}_deliverables.yaml"
            with open(case_deliverables_path, "w") as fh:
                yaml.dump({"files": []}, fh)
            deliverables_paths[case] = case_deliverables_path

    return deliverables_paths


@pytest.fixture(name="mip_qc_sample_info")
def fixture_mip_qc_sample_info(mip_case_ids: dict, mip_case_dirs: dict) -> dict:
    """Create qc sample info files for cases"""

    yaml = YAML()

    # Store paths in a list to be used in other stages

    qc_sample_info = {}

    for case in mip_case_ids:
        sample_info_data = {
            "date": dt.datetime.now(),
            "analysisrunstatus": "finished",
            "analysis_date": dt.datetime.now(),
            "case": case,
            "mip_version": "latest",
        }

        sample_info_path = mip_case_dirs[case] / f"{case}_qc_sample_info.yaml"
        qc_sample_info[case] = sample_info_path
        with open(sample_info_path, "w") as fh:
            yaml.dump(sample_info_data, fh)

    return qc_sample_info


@pytest.fixture(name="mip_configs")
def fixture_mip_configs(
    mip_case_ids: dict, mip_case_dirs: dict, mip_deliverables: dict, mip_qc_sample_info: dict
) -> dict:
    """Create config files for mip case"""

    yaml = YAML()

    # Storing paths in a list to be returned and used in later stages.

    config_dict = {}

    for case in mip_case_ids:
        config = {
            "email": "fake_email@notscilifelab.com",
            "case_id": case,
            "samples": "dummy_samples",
            "is_dryrun": False,
            "outdata_dir": str(mip_case_dirs[case]),
            "priority": "medium",
            "sampleinfo_path": str(mip_qc_sample_info[case]),
            "sample_info_file": str(mip_qc_sample_info[case]),
            "analysis_type": {mip_case_ids[case]["internal_id"]: "wgs"},
            "slurm_quality_of_service": "medium",
            "store_file": str(mip_deliverables[case]),
        }
        analysis_path = Path(mip_case_dirs[case] / "analysis")
        Path(analysis_path).mkdir(parents=True, exist_ok=True)
        config_path = analysis_path / f"{case}_config.yaml"
        config_dict[case] = config_path
        with open(config_path, "w") as fh:
            yaml.dump(config, fh)

    return config_dict


@pytest.fixture(name="_store")
def fixture_store(base_store: Store, mip_case_ids: dict, helpers) -> Store:
    """Create and populate temporary db with test cases for mip"""

    _store = base_store

    # Add apptag to db

    helpers.ensure_application_version(store=_store, application_tag="WGSA", application_type="wgs")

    # Add sample, cases and relationships to db

    for case in mip_case_ids:
        family = helpers.add_family(
            store=_store,
            internal_id=case,
            family_id=mip_case_ids[case]["name"],
        )
        sample = helpers.add_sample(
            store=_store,
            sample=mip_case_ids[case]["internal_id"],
            data_analysis=Pipeline.MIP_DNA,
            customer_name="cust000",
            application_tag="WGSA",
            application_type="wgs",
            gender="unknown",
        )
        helpers.add_relationship(store=_store, sample=sample, family=family, status="affected")

    return _store


@pytest.fixture(name="empty_housekeeper_api")
def fixture_empty_housekeeper_api(real_housekeeper_api: HousekeeperAPI) -> HousekeeperAPI:
    """Empty housekeeper for testing store in mip workflow"""

    _housekeeper_api = real_housekeeper_api

    return _housekeeper_api


@pytest.fixture(name="populated_mip_tb_api")
def fixture_populated_mip_tb_api(
    trailblazer_api: TrailblazerAPI, mip_case_ids: dict, mip_configs: dict
) -> TrailblazerAPI:
    """Trailblazer api filled with mip cases"""

    _tb_api = trailblazer_api

    for case in mip_case_ids:
        _tb_api.add_commit(
            tb_Analysis(
                family=case, status="completed", deleted=False, config_path=str(mip_configs[case])
            )
        )

    return _tb_api


@pytest.fixture
def mip_context(
    analysis_store_single_case, trailblazer_api, housekeeper_api, mip_lims, mock_root_folder
):
    return {
        "dna_api": MipAnalysisAPI(
            db=analysis_store_single_case,
            hk_api=housekeeper_api,
            tb_api=trailblazer_api,
            scout_api="scout_api",
            lims_api=mip_lims,
            script="echo",
            pipeline="analyse rd_dna",
            conda_env="S_mip_rd-dna",
            root=mock_root_folder,
        ),
        "rna_api": MipAnalysisAPI(
            db=analysis_store_single_case,
            hk_api=housekeeper_api,
            tb_api=trailblazer_api,
            scout_api="scout_api",
            lims_api=mip_lims,
            script="echo",
            pipeline="analyse rd_rna",
            conda_env="S_mip_rd-rna",
            root=mock_root_folder,
        ),
        "mip-rd-dna": {
            "conda_env": "S_mip_rd-dna",
            "mip_config": "config.yaml",
            "pipeline": "analyse rd_dna",
            "root": mock_root_folder,
            "script": "echo",
        },
        "mip-rd-rna": {
            "conda_env": "S_mip_rd-rna",
            "mip_config": "config.yaml",
            "pipeline": "analyse rd_rna",
            "root": mock_root_folder,
            "script": "echo",
        },
    }


@pytest.fixture(name="mip_store_context")
def mip_store_context(
    trailblazer_api, _store: Store, empty_housekeeper_api: HousekeeperAPI, mock_root_folder: Path
) -> dict:
    """Create a context to be used in testing mip store, this should be fused with mip_context above at later stages"""
    return {
        "mip_api": MipAnalysisAPI(
            db=_store,
            tb_api=trailblazer_api,
            hk_api=empty_housekeeper_api,
            scout_api=None,
            lims_api=None,
            script="None",
            pipeline="None",
            conda_env="None",
            root=mock_root_folder,
        ),
        "trailblazer_api": trailblazer_api,
        "housekeeper_api": empty_housekeeper_api,
        "status_db": _store,
    }



from typing import List

import xlrd
from cg.constants import METAGENOME_SOURCES, ANALYSIS_SOURCES, Pipeline

from cg.exc import OrderFormError
from cg.meta.orders import OrderType
from cg.utils.StrEnum import StrEnum

SEX_MAP = {"male": "M", "female": "F", "unknown": "unknown"}
REV_SEX_MAP = {value: key for key, value in SEX_MAP.items()}
CONTAINER_TYPES = ["Tube", "96 well plate"]
SOURCE_TYPES = set().union(METAGENOME_SOURCES, ANALYSIS_SOURCES)
VALID_ORDERFORMS = [
    "1508:21",  # Orderform MIP, Balsamic, sequencing only, MIP RNA
    "1541:6",  # Orderform Externally sequenced samples
    "1603:9",  # Microbial WGS
    "1604:9",  # Orderform Ready made libraries (RML)
    "1605:8",  # Microbial metagenomes
]


CASE_PROJECT_TYPES = [
    str(OrderType.MIP_DNA),
    str(OrderType.EXTERNAL),
    str(OrderType.BALSAMIC),
    str(OrderType.MIP_RNA),
]


def check_orderform_version(document_title):
    """Raise an error if the orderform is too new or too old for the order portal."""
    for valid_orderform in VALID_ORDERFORMS:
        if valid_orderform in document_title:
            return
    raise OrderFormError(f"Unsupported orderform: {document_title}")


def parse_orderform(excel_path: str) -> dict:
    """Parse out information from an order form."""
    workbook = xlrd.open_workbook(excel_path)

    sheet_name = None
    sheet_names = workbook.sheet_names()
    for name in ["orderform", "order form"]:
        if name in sheet_names:
            sheet_name = name
            break
    if sheet_name is None:
        raise OrderFormError("'orderform' sheet not found in Excel file")
    orderform_sheet = workbook.sheet_by_name(sheet_name)

    document_title = get_document_title(workbook, orderform_sheet)
    check_orderform_version(document_title)

    raw_samples = relevant_rows(orderform_sheet)
    if len(raw_samples) == 0:
        raise OrderFormError("orderform doesn't contain any samples")
    parsed_samples = [parse_sample(raw_sample) for raw_sample in raw_samples]

    project_type = get_project_type(document_title, parsed_samples)

    if project_type in CASE_PROJECT_TYPES:
        parsed_cases = group_cases(parsed_samples)
        items = []
        customer_ids = set()
        for case_id, parsed_case in parsed_cases.items():
            customer_id, case_data = expand_case(case_id, parsed_case)
            customer_ids.add(customer_id)
            items.append(case_data)
    else:
        customer_ids = set(sample["customer"] for sample in parsed_samples)
        items = parsed_samples

    customer_options = len(customer_ids)
    if customer_options == 0:
        raise OrderFormError("Customer information is missing")
    elif customer_options != 1:
        raise OrderFormError(f"Samples have different customers: {customer_ids}")

    data = {"customer": customer_ids.pop(), "items": items, "project_type": project_type}

    return data


def get_document_title(workbook: xlrd.book.Book, orderform_sheet: xlrd.sheet.Sheet) -> str:
    """Get the document title for the order form."""
    if "information" in workbook.sheet_names():
        information_sheet = workbook.sheet_by_name("information")
        document_title = information_sheet.row(0)[2].value
        return document_title

    document_title = orderform_sheet.row(0)[1].value
    return document_title


def get_project_type(document_title: str, parsed_samples: List) -> str:
    """Determine the project type."""

    project_type = None

    if "1541" in document_title:
        project_type = "external"
    elif "1604" in document_title:
        project_type = "rml"
    elif "1603" in document_title:
        project_type = "microsalt"
    elif "1605" in document_title:
        project_type = "metagenome"
    elif "1508" in document_title:
        analyses = set(sample["data_analysis"].lower() for sample in parsed_samples)

        if len(analyses) != 1:
            raise OrderFormError(f"mixed 'Data Analysis' types: {', '.join(analyses)}")

        project_type = analyses.pop()
        if "mip" in project_type and "balsamic" in project_type:
            raise OrderFormError(f"mixed 'Data Analysis' types: {project_type}")

    return project_type


def expand_case(case_id, parsed_case):
    """Fill-in information about families."""
    new_case = {"name": case_id, "samples": []}
    samples = parsed_case["samples"]

    require_qcoks = set(raw_sample["require_qcok"] for raw_sample in samples)
    new_case["require_qcok"] = True in require_qcoks

    priorities = set(raw_sample["priority"] for raw_sample in samples)
    if len(priorities) == 1:
        new_case["priority"] = priorities.pop()
    else:
        raise OrderFormError(f"multiple values for 'Priority' for case: {case_id}")

    customers = set(raw_sample["customer"] for raw_sample in samples)
    if len(customers) != 1:
        raise OrderFormError("Invalid customer information: {}".format(customers))
    customer = customers.pop()
    gene_panels = set()
    for raw_sample in samples:
        if raw_sample["panels"]:
            gene_panels.update(raw_sample["panels"])
        new_sample = {
            "name": raw_sample["name"],
            "sex": raw_sample["sex"],
            "application": raw_sample["application"],
            "source": raw_sample["source"],
        }
        if raw_sample.get("container") in CONTAINER_TYPES:
            new_sample["container"] = raw_sample["container"]

        for key in (
            "capture_kit",
            "comment",
            "container_name",
            "data_analysis",
            "elution_buffer",
            "formalin_fixation_time",
            "from_sample",
            "post_formalin_fixation_time",
            "quantity",
            "status",
            "time_point",
            "tissue_block_size",
            "tumour",
            "tumour_purity",
            "well_position",
        ):
            if raw_sample.get(key):
                new_sample[key] = raw_sample[key]

        for parent_id in ("mother", "father"):
            if raw_sample[parent_id]:
                new_sample[parent_id] = raw_sample[parent_id]
        new_case["samples"].append(new_sample)

    new_case["panels"] = list(gene_panels)

    return customer, new_case


def group_cases(parsed_samples):
    """Group samples on case."""
    raw_cases = {}
    for sample in parsed_samples:
        case_id = sample["case"]
        if case_id not in raw_cases:
            raw_cases[case_id] = {"samples": []}
        raw_cases[case_id]["samples"].append(sample)
    return raw_cases


def parse_sample(raw_sample):
    """Parse a raw sample row from order form sheet."""
    if ":" in raw_sample.get("UDF/Gene List", ""):
        raw_sample["UDF/Gene List"] = raw_sample["UDF/Gene List"].replace(":", ";")

    if raw_sample["UDF/priority"].lower() == "förtur":
        raw_sample["UDF/priority"] = "priority"
    raw_source = raw_sample.get("UDF/Source")
    sample = {
        "application": raw_sample["UDF/Sequencing Analysis"],
        "capture_kit": raw_sample.get("UDF/Capture Library version"),
        "case": raw_sample.get("UDF/familyID"),
        "comment": raw_sample.get("UDF/Comment"),
        "container": raw_sample.get("Container/Type"),
        "container_name": raw_sample.get("Container/Name"),
        "custom_index": raw_sample.get("UDF/Custom index"),
        "customer": raw_sample["UDF/customer"],
        "data_delivery": raw_sample.get("UDF/Data Delivery"),
        "elution_buffer": raw_sample.get("UDF/Sample Buffer"),
        "extraction_method": raw_sample.get("UDF/Extraction method"),
        "formalin_fixation_time": raw_sample.get("UDF/Formalin Fixation Time"),
        "index": raw_sample.get("UDF/Index type"),
        "from_sample": raw_sample.get("UDF/is_for_sample"),
        "name": raw_sample["Sample/Name"],
        "organism": raw_sample.get("UDF/Strain"),
        "organism_other": raw_sample.get("UDF/Other species"),
        "panels": (
            raw_sample["UDF/Gene List"].split(";") if raw_sample.get("UDF/Gene List") else None
        ),
        "pool": raw_sample.get("UDF/pool name"),
        "post_formalin_fixation_time": raw_sample.get("UDF/Post Formalin Fixation Time"),
        "priority": raw_sample["UDF/priority"].lower() if raw_sample.get("UDF/priority") else None,
        "reagent_label": raw_sample.get("Sample/Reagent Label"),
        "reference_genome": raw_sample.get("UDF/Reference Genome Microbial"),
        "require_qcok": raw_sample.get("UDF/Process only if QC OK") == "yes",
        "rml_plate_name": raw_sample.get("UDF/RML plate name"),
        "sex": REV_SEX_MAP.get(raw_sample.get("UDF/Gender", "").strip()),
        "source": raw_source if raw_source in SOURCE_TYPES else None,
        "status": raw_sample["UDF/Status"].lower() if raw_sample.get("UDF/Status") else None,
        "tissue_block_size": raw_sample.get("UDF/Tissue Block Size"),
        "tumour": raw_sample.get("UDF/tumor") == "yes",
        "tumour_purity": raw_sample.get("UDF/tumour purity"),
        "well_position": raw_sample.get("Sample/Well Location"),
        "well_position_rml": raw_sample.get("UDF/RML well position"),
    }

    data_analysis = raw_sample.get("UDF/Data Analysis").lower()

    if data_analysis and "balsamic" in data_analysis:
        sample["data_analysis"] = str(Pipeline.BALSAMIC)
    elif data_analysis and "rna" in data_analysis:
        sample["data_analysis"] = str(Pipeline.MIP_RNA)
    elif data_analysis and "mip" in data_analysis or "scout" in data_analysis:
        sample["data_analysis"] = str(Pipeline.MIP_DNA)
    elif data_analysis and "microbial" in data_analysis:
        sample["data_analysis"] = str(Pipeline.MICROSALT)
    elif data_analysis and ("fastq" in data_analysis or "custom" in data_analysis):
        sample["data_analysis"] = str(Pipeline.FASTQ)
    else:
        raise OrderFormError(f"unknown 'Data Analysis' for order: {data_analysis}")

    numeric_values = [
        ("index_number", "UDF/Index number"),
        ("volume", "UDF/Volume (uL)"),
        ("quantity", "UDF/Quantity"),
        ("concentration", "UDF/Concentration (nM)"),
        ("concentration_weight", "UDF/Sample Conc."),
        ("time_point", "UDF/time_point"),
    ]
    for json_key, excel_key in numeric_values:
        str_value = raw_sample.get(excel_key, "").rsplit(".0")[0]
        if str_value.replace(".", "").isnumeric():
            sample[json_key] = str_value

    for parent in ["mother", "father"]:
        parent_key = f"UDF/{parent}ID"
        sample[parent] = (
            raw_sample[parent_key]
            if raw_sample.get(parent_key) and (raw_sample[parent_key] != "0.0")
            else None
        )

    return sample


def relevant_rows(orderform_sheet):
    """Get the relevant rows from an order form sheet."""
    raw_samples = []
    current_row = None
    empty_row_found = False
    for row in orderform_sheet.get_rows():
        if row[0].value == "</SAMPLE ENTRIES>":
            break

        if current_row == "header":
            header_row = [cell.value for cell in row]
            current_row = None
        elif current_row == "samples":
            values = [str(cell.value) for cell in row]

            # skip empty rows
            if values[0]:
                if empty_row_found:
                    raise OrderFormError(
                        f"Found data after empty lines. Please delete any "
                        f"non-sample data rows in between the samples"
                    )

                sample_dict = dict(zip(header_row, values))

                raw_samples.append(sample_dict)
            else:
                empty_row_found = True

        if row[0].value == "<TABLE HEADER>":
            current_row = "header"
        elif row[0].value == "<SAMPLE ENTRIES>":
            current_row = "samples"
    return raw_samples



"""Functions to get cases from the store database"""

from typing import Iterator

from sqlalchemy.orm import Query

from cg.store import models
from cg.store.api import Store


def analysis_completed(store: Store) -> Query:
    """Return cases where analysis is finished"""
    cases = (
        store.Family.query.join(models.Analysis)
        .filter(models.Family.analyses)
        .filter(models.Family.action == None)
        .order_by(models.Family.created_at.asc())
    )
    return cases


def ready_for_spring_compression(store: Store) -> Iterator[models.Family]:
    """Fetch all cases that are ready to be compressed by SPRING"""
    for case_obj in analysis_completed(store):
        if case_obj.ordered_at > case_obj.analyses[0].completed_at:
            continue
        yield case_obj



"""Fixtures for testing balsamic app"""
import string

import pytest


@pytest.fixture
def valid_fastq_filename_pattern():
    """the pattern usalt file names should match"""
    # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz

    # 'sample_flowcell_lane_read.fastq.gz'
    return r"^ACC.+_.+_L[1-8]{1}_[1-2]{1}\.fastq\.gz$"


def _full_content():
    """The content the files are made of"""
    return string.ascii_letters


def simple(tmpdir):
    """Creates a dict with the data to use in the tests"""
    flowcells = [1]
    lanes = [1, 2, 3]
    reads = [1, 2]

    _simple = {
        "files": [],
        "content_r1": [],
        "content_r2": [],
        "data": [],
        "data_reversed": [],
    }
    i = 0

    for read in reads:
        for flowcell in flowcells:
            for lane in lanes:
                content = _full_content()[i]
                file_path = create_file(tmpdir, flowcell, lane, read, content)

                _simple["files"].append(file_path)

                if read == 1:
                    _simple["content_r1"].append(content)
                else:
                    _simple["content_r2"].append(content)

                data = create_file_data(file_path, flowcell, lane, read)
                _simple["data"].append(data)
                _simple["data_reversed"].insert(0, data)
                i += 1

    return _simple


@pytest.fixture
def simple_files_data(tmpdir):
    """Data for link method"""
    return simple(tmpdir)["data"]


def create_file(tmpdir, flowcell, lane, read, file_content):
    """actual file on disk"""

    # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz sample_flowcell_lane_read.fastq.gz
    file_name = f"ACC123_FC000{flowcell}_L{lane}_{read}.fastq.gz"
    file_path = tmpdir / file_name
    file_path.write(file_content)
    return file_path


def create_file_data(file_path, flowcell, lane, read):
    """meta data about a file on disk"""
    data = {
        "path": file_path,
        "lane": lane,
        "flowcell": flowcell,
        "read": read,
        "undetermined": False,
    }
    return data


@pytest.fixture
def cg_config(tmpdir):
    """mock relevant parts of a cg-config"""
    return {"microsalt": {"root": tmpdir}}


@pytest.fixture
def link_ticket():
    """mock ticket number"""
    return "123456"


@pytest.fixture
def link_sample():
    """mock sample name"""
    return "sample"



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_dry_order 1"
] = """[
    {
        "CG_ID_project": "microbial_order_test",
        "CG_ID_sample": "microbial_sample_id",
        "Customer_ID": "cust_test",
        "Customer_ID_project": 123456,
        "Customer_ID_sample": "microbial_name_test",
        "application_tag": "dummy_tag",
        "date_arrival": "0001-01-01 00:00:00",
        "date_libprep": "0001-01-01 00:00:00",
        "date_sequencing": "0001-01-01 00:00:00",
        "method_libprep": "1337:00",
        "method_sequencing": "1338:00",
        "organism": "organism_test",
        "priority": "research",
        "reference": null
    }
]
"""

snapshots[
    "test_dry_sample 1"
] = """[
    {
        "CG_ID_project": "microbial_order_test",
        "CG_ID_sample": "microbial_sample_id",
        "Customer_ID": "cust_test",
        "Customer_ID_project": 123456,
        "Customer_ID_sample": "microbial_name_test",
        "application_tag": "dummy_tag",
        "date_arrival": "0001-01-01 00:00:00",
        "date_libprep": "0001-01-01 00:00:00",
        "date_sequencing": "0001-01-01 00:00:00",
        "method_libprep": "1337:00",
        "method_sequencing": "1338:00",
        "organism": "organism_test",
        "priority": "research",
        "reference": null
    }
]
"""

snapshots[
    "test_dry_sample_order 1"
] = """[
    {
        "CG_ID_project": "microbial_order_test",
        "CG_ID_sample": "microbial_sample_id",
        "Customer_ID": "cust_test",
        "Customer_ID_project": 123456,
        "Customer_ID_sample": "microbial_name_test",
        "application_tag": "dummy_tag",
        "date_arrival": "0001-01-01 00:00:00",
        "date_libprep": "0001-01-01 00:00:00",
        "date_sequencing": "0001-01-01 00:00:00",
        "method_libprep": "1337:00",
        "method_sequencing": "1338:00",
        "organism": "organism_test",
        "priority": "research",
        "reference": null
    }
]
"""

snapshots["test_sample 1"] = [
    """[
""",
    """    {
""",
    """        "CG_ID_project": "microbial_order_test",
""",
    """        "CG_ID_sample": "microbial_sample_id",
""",
    """        "Customer_ID": "cust_test",
""",
    """        "Customer_ID_project": 123456,
""",
    """        "Customer_ID_sample": "microbial_name_test",
""",
    """        "application_tag": "dummy_tag",
""",
    """        "date_arrival": "0001-01-01 00:00:00",
""",
    """        "date_libprep": "0001-01-01 00:00:00",
""",
    """        "date_sequencing": "0001-01-01 00:00:00",
""",
    """        "method_libprep": "1337:00",
""",
    """        "method_sequencing": "1338:00",
""",
    """        "organism": "organism_test",
""",
    """        "priority": "research",
""",
    """        "reference": null
""",
    """    }
""",
    "]",
]
